"""GQL implementation of mutation tests.

Tests CRUD operations using GQL (ISO standard) query language.
"""

from tests.bases.test_mutations import BaseMutationsTest


class TestGQLMutations(BaseMutationsTest):
    """GQL implementation of mutation tests."""

    def create_node_query(self, labels: list[str], props: dict) -> str:
        """GQL: INSERT (:<labels> {<props>})"""
        label_str = ":".join(labels) if labels else ""
        if label_str:
            label_str = f":{label_str}"

        prop_parts = []
        for k, v in props.items():
            if isinstance(v, str):
                prop_parts.append(f"{k}: '{v}'")
            elif isinstance(v, bool):
                prop_parts.append(f"{k}: {'true' if v else 'false'}")
            elif v is None:
                prop_parts.append(f"{k}: null")
            else:
                prop_parts.append(f"{k}: {v}")

        prop_str = ", ".join(prop_parts)
        return f"INSERT (n{label_str} {{{prop_str}}}) RETURN n"

    def match_node_query(self, label: str, return_prop: str = "name") -> str:
        """GQL: MATCH (n:<label>) RETURN n.<prop>"""
        return f"MATCH (n:{label}) RETURN n.{return_prop}"

    def match_where_query(
        self, label: str, prop: str, op: str, value, return_prop: str = "name"
    ) -> str:
        """GQL: MATCH (n:<label>) WHERE n.<prop> <op> <value> RETURN n.<return_prop>"""
        if isinstance(value, str):
            value_str = f"'{value}'"
        else:
            value_str = str(value)
        return f"MATCH (n:{label}) WHERE n.{prop} {op} {value_str} RETURN n.{return_prop}"

    def delete_node_query(self, label: str, prop: str, value) -> str:
        """GQL: MATCH (n:<label>) WHERE n.<prop> = <value> DELETE n"""
        if isinstance(value, str):
            value_str = f"'{value}'"
        else:
            value_str = str(value)
        return f"MATCH (n:{label}) WHERE n.{prop} = {value_str} DELETE n"

    def create_edge_query(
        self,
        from_label: str,
        from_prop: str,
        from_value,
        to_label: str,
        to_prop: str,
        to_value,
        edge_type: str,
        edge_props: dict,
    ) -> str:
        """GQL: MATCH (a:<from_label>), (b:<to_label>) WHERE ... CREATE (a)-[:<edge_type>]->(b)"""
        from_val = f"'{from_value}'" if isinstance(from_value, str) else from_value
        to_val = f"'{to_value}'" if isinstance(to_value, str) else to_value

        prop_parts = []
        for k, v in edge_props.items():
            if isinstance(v, str):
                prop_parts.append(f"{k}: '{v}'")
            else:
                prop_parts.append(f"{k}: {v}")
        prop_str = ", ".join(prop_parts) if prop_parts else ""

        if prop_str:
            return (
                f"MATCH (a:{from_label}), (b:{to_label}) "
                f"WHERE a.{from_prop} = {from_val} AND b.{to_prop} = {to_val} "
                f"CREATE (a)-[r:{edge_type} {{{prop_str}}}]->(b) RETURN r"
            )
        else:
            return (
                f"MATCH (a:{from_label}), (b:{to_label}) "
                f"WHERE a.{from_prop} = {from_val} AND b.{to_prop} = {to_val} "
                f"CREATE (a)-[r:{edge_type}]->(b) RETURN r"
            )

    def update_node_query(
        self, label: str, match_prop: str, match_value, set_prop: str, set_value
    ) -> str:
        """GQL: MATCH (n:<label>) WHERE n.<prop> = <value> SET n.<prop> = <value>"""
        match_val = f"'{match_value}'" if isinstance(match_value, str) else match_value
        set_val = f"'{set_value}'" if isinstance(set_value, str) else set_value
        return (
            f"MATCH (n:{label}) WHERE n.{match_prop} = {match_val} "
            f"SET n.{set_prop} = {set_val} RETURN n"
        )


# =============================================================================
# GQL-SPECIFIC MUTATION TESTS
# =============================================================================


class TestGQLSpecificMutations:
    """GQL-specific mutation tests."""

    def test_gql_insert_syntax(self, db):
        """Test GQL INSERT syntax specifically."""
        result = db.execute("INSERT (:Person {name: 'InsertTest', age: 42}) RETURN *")
        rows = list(result)

        result = db.execute("MATCH (n:Person) WHERE n.name = 'InsertTest' RETURN n.age")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.age"] == 42

    def test_gql_multiple_labels_syntax(self, db):
        """Test GQL multiple labels syntax."""
        result = db.execute("INSERT (:Person:Developer:Senior {name: 'MultiLabel'}) RETURN *")
        list(result)

        result = db.execute("MATCH (n:Person:Developer) RETURN n.name")
        rows = list(result)
        assert len(rows) >= 1

    def test_gql_property_types(self, db):
        """Test various property types in GQL."""
        db.execute("INSERT (:Data {str: 'hello', num: 42, flt: 3.14, bool: true})")

        result = db.execute("MATCH (n:Data) RETURN n.str, n.num, n.flt, n.bool")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.str"] == "hello"
        assert rows[0]["n.num"] == 42

    def test_gql_set_multiple_properties(self, db):
        """Test SET with multiple properties."""
        db.execute("INSERT (:Person {name: 'Alix', age: 30, city: 'NYC'})")

        db.execute("MATCH (n:Person) WHERE n.name = 'Alix' SET n.age = 31, n.city = 'LA' RETURN n")

        result = db.execute("MATCH (n:Person) WHERE n.name = 'Alix' RETURN n.age, n.city")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.age"] == 31
        assert rows[0]["n.city"] == "LA"

    def test_gql_remove_property(self, db):
        """Test removing property by setting to null (REMOVE not yet in GQL)."""
        db.execute("INSERT (:Person {name: 'Gus', age: 25, temp: 'delete_me'})")

        # GQL uses SET n.prop = null to remove properties (REMOVE not yet implemented)
        db.execute("MATCH (n:Person) WHERE n.name = 'Gus' SET n.temp = null RETURN n")

        result = db.execute("MATCH (n:Person) WHERE n.name = 'Gus' RETURN n.temp")
        rows = list(result)
        # Property should be null or not present
        if len(rows) > 0:
            assert rows[0].get("n.temp") is None

    def test_gql_detach_delete(self, db):
        """Test DETACH DELETE (delete node and all relationships)."""
        alix = db.create_node(["Person"], {"name": "Alix"})
        gus = db.create_node(["Person"], {"name": "Gus"})
        db.create_edge(alix.id, gus.id, "KNOWS", {})

        # Delete Alix and her relationships
        db.execute("MATCH (n:Person) WHERE n.name = 'Alix' DETACH DELETE n")

        # Verify Alix is deleted
        result = db.execute("MATCH (n:Person) RETURN n.name")
        rows = list(result)
        names = [r["n.name"] for r in rows]
        assert "Alix" not in names
        assert "Gus" in names

    def test_gql_merge_create(self, db):
        """Test MERGE creates node if not exists."""
        db.execute("MERGE (:Person {name: 'MergeTest'})")

        result = db.execute("MATCH (n:Person) WHERE n.name = 'MergeTest' RETURN n.name")
        rows = list(result)
        assert len(rows) == 1

    def test_gql_merge_match(self, db):
        """Test MERGE matches existing node."""
        db.execute("INSERT (:Person {name: 'MergeExisting', age: 30})")

        # MERGE should match existing node, not create new
        db.execute("MERGE (n:Person {name: 'MergeExisting'}) SET n.age = 31 RETURN n")

        result = db.execute(
            "MATCH (n:Person) WHERE n.name = 'MergeExisting' RETURN count(n) AS cnt"
        )
        rows = list(result)
        assert rows[0]["cnt"] == 1

        result = db.execute("MATCH (n:Person) WHERE n.name = 'MergeExisting' RETURN n.age")
        rows = list(result)
        assert rows[0]["n.age"] == 31

    def test_gql_insert_negative_float(self, db):
        """Test INSERT with negative float property (issue #160)."""
        db.execute("INSERT (:Point {lat: -6.248, lon: 106.845})")

        result = db.execute("MATCH (n:Point) RETURN n.lat, n.lon")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.lat"] == -6.248
        assert rows[0]["n.lon"] == 106.845

    def test_gql_insert_negative_integer(self, db):
        """Test INSERT with negative integer property (issue #160)."""
        db.execute("INSERT (:Floor {level: -3})")

        result = db.execute("MATCH (n:Floor) RETURN n.level")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.level"] == -3

    def test_gql_insert_multiple_negative_properties(self, db):
        """Test INSERT with multiple negative properties (issue #160)."""
        db.execute("INSERT (:Point {lat: -6.248, lon: -106.845, alt: -42})")

        result = db.execute("MATCH (n:Point) RETURN n.lat, n.lon, n.alt")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.lat"] == -6.248
        assert rows[0]["n.lon"] == -106.845
        assert rows[0]["n.alt"] == -42

    def test_gql_insert_geographic_coordinates(self, db):
        """Test realistic lat/lon storage, the original issue #160 use case."""
        db.execute("INSERT (:City {name: 'Jakarta', lat: -6.175, lon: 106.827})")
        db.execute("INSERT (:City {name: 'Buenos Aires', lat: -34.604, lon: -58.382})")

        result = db.execute("MATCH (n:City {name: 'Buenos Aires'}) RETURN n.lat, n.lon")
        rows = list(result)
        assert rows[0]["n.lat"] == -34.604
        assert rows[0]["n.lon"] == -58.382

    def test_gql_negative_edge_property_inline(self, db):
        """Test negative property on an inline edge (issue #160)."""
        db.execute("INSERT (:Node {name: 'Alix'})-[:LINK {weight: -0.5}]->(:Node {name: 'Gus'})")

        result = db.execute("MATCH ()-[r:LINK]->() RETURN r.weight")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["r.weight"] == -0.5

    def test_gql_negative_edge_property_via_match(self, db):
        """Test negative property on edge created via MATCH/CREATE (issue #160)."""
        db.execute("INSERT (:Node {name: 'Alix'})")
        db.execute("INSERT (:Node {name: 'Gus'})")
        db.execute(
            "MATCH (a:Node {name: 'Alix'}), (b:Node {name: 'Gus'}) "
            "CREATE (a)-[:SCORED {score: -10}]->(b)"
        )

        result = db.execute("MATCH ()-[r:SCORED]->() RETURN r.score")
        rows = list(result)
        assert rows[0]["r.score"] == -10

    def test_gql_set_negative_integer(self, db):
        """Test SET with negative integer (issue #160)."""
        db.execute("INSERT (:Floor {name: 'Parking', level: 1})")
        db.execute("MATCH (n:Floor {name: 'Parking'}) SET n.level = -2")

        result = db.execute("MATCH (n:Floor {name: 'Parking'}) RETURN n.level")
        rows = list(result)
        assert rows[0]["n.level"] == -2

    def test_gql_set_negative_float(self, db):
        """Test SET with negative float (issue #160)."""
        db.execute("INSERT (:Sensor {name: 'Temp', reading: 25.0})")
        db.execute("MATCH (n:Sensor {name: 'Temp'}) SET n.reading = -18.5")

        result = db.execute("MATCH (n:Sensor {name: 'Temp'}) RETURN n.reading")
        rows = list(result)
        assert rows[0]["n.reading"] == -18.5

    def test_gql_where_equals_negative(self, db):
        """Test WHERE with negative equality (issue #160)."""
        db.execute("INSERT (:Floor {name: 'B1', level: -1})")
        db.execute("INSERT (:Floor {name: 'B2', level: -2})")
        db.execute("INSERT (:Floor {name: 'G', level: 0})")

        result = db.execute("MATCH (n:Floor) WHERE n.level = -2 RETURN n.name")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.name"] == "B2"

    def test_gql_where_less_than_negative(self, db):
        """Test WHERE with less-than negative comparison (issue #160)."""
        db.execute("INSERT (:Floor {name: 'B1', level: -1})")
        db.execute("INSERT (:Floor {name: 'B2', level: -2})")
        db.execute("INSERT (:Floor {name: 'B3', level: -3})")

        result = db.execute("MATCH (n:Floor) WHERE n.level < -1 RETURN n.name ORDER BY n.level")
        rows = list(result)
        assert len(rows) == 2

    def test_gql_where_between_negatives(self, db):
        """Test WHERE filtering between two negative values (issue #160)."""
        for val in [-10, -5, -1, 0, 5]:
            db.execute(f"INSERT (:Data {{val: {val}}})")

        result = db.execute("MATCH (n:Data) WHERE n.val >= -5 AND n.val <= -1 RETURN n.val")
        rows = list(result)
        assert len(rows) == 2

    def test_gql_property_map_match_negative(self, db):
        """Test inline property map matching with negative value (issue #160)."""
        db.execute("INSERT (:Floor {level: -1, name: 'Basement'})")
        db.execute("INSERT (:Floor {level: 1, name: 'Ground'})")

        result = db.execute("MATCH (n:Floor {level: -1}) RETURN n.name")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.name"] == "Basement"

    def test_gql_merge_create_with_negative(self, db):
        """Test MERGE creating node with negative property (issue #160)."""
        db.execute("MERGE (:Floor {level: -1, name: 'Basement'})")

        result = db.execute("MATCH (n:Floor) RETURN n.level, n.name")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.level"] == -1
        assert rows[0]["n.name"] == "Basement"

    def test_gql_merge_match_with_negative(self, db):
        """Test MERGE matching existing node with negative property (issue #160)."""
        db.execute("INSERT (:Floor {level: -1, name: 'Basement'})")
        db.execute("MERGE (n:Floor {level: -1}) ON MATCH SET n.name = 'SubBasement'")

        result = db.execute("MATCH (n:Floor) RETURN n.level, n.name")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.level"] == -1
        assert rows[0]["n.name"] == "SubBasement"

    def test_gql_merge_on_create_set_negative(self, db):
        """Test MERGE ON CREATE SET with negative value (issue #160)."""
        db.execute("MERGE (n:Sensor {name: 'Temp'}) ON CREATE SET n.reading = -40")

        result = db.execute("MATCH (n:Sensor {name: 'Temp'}) RETURN n.reading")
        rows = list(result)
        assert rows[0]["n.reading"] == -40

    def test_gql_merge_on_match_set_negative(self, db):
        """Test MERGE ON MATCH SET with negative value (issue #160)."""
        db.execute("INSERT (:Sensor {name: 'Temp', reading: 25})")
        db.execute("MERGE (n:Sensor {name: 'Temp'}) ON MATCH SET n.reading = -18")

        result = db.execute("MATCH (n:Sensor {name: 'Temp'}) RETURN n.reading")
        rows = list(result)
        assert rows[0]["n.reading"] == -18

    def test_gql_return_negative_literal(self, db):
        """Test RETURN with negative literal (issue #160)."""
        result = db.execute("RETURN -5 AS val")
        rows = list(result)
        assert rows[0]["val"] == -5

    def test_gql_negative_in_arithmetic(self, db):
        """Test negative value in arithmetic expression (issue #160)."""
        db.execute("INSERT (:Data {val: 10})")

        result = db.execute("MATCH (n:Data) RETURN n.val + -5 AS result")
        rows = list(result)
        assert rows[0]["result"] == 5



"""GQL spec: Functions (ISO sec 20).

Covers: aggregation, string, numeric, temporal, list, graph element, path,
type conversion functions.
"""

import math

# =============================================================================
# Aggregation Functions
# =============================================================================


class TestAggregationFunctions:
    """COUNT, SUM, AVG, MIN, MAX, COLLECT, STDEV, PERCENTILE, LISTAGG."""

    def test_count_star(self, db):
        """COUNT(n) counts all rows."""
        db.create_node(["N"], {"v": 1})
        db.create_node(["N"], {"v": 2})
        result = list(db.execute("MATCH (n:N) RETURN count(n) AS cnt"))
        assert result[0]["cnt"] == 2

    def test_count_expr(self, db):
        """COUNT(expr) counts non-null values."""
        db.create_node(["N"], {"v": 1})
        db.create_node(["N"], {})
        result = list(db.execute("MATCH (n:N) RETURN count(n.v) AS cnt"))
        assert result[0]["cnt"] == 1

    def test_count_distinct(self, db):
        """COUNT(DISTINCT expr) counts unique values."""
        db.create_node(["N"], {"v": "a"})
        db.create_node(["N"], {"v": "a"})
        db.create_node(["N"], {"v": "b"})
        result = list(db.execute("MATCH (n:N) RETURN count(DISTINCT n.v) AS cnt"))
        assert result[0]["cnt"] == 2

    def test_sum(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 20})
        db.create_node(["N"], {"v": 30})
        result = list(db.execute("MATCH (n:N) RETURN sum(n.v) AS s"))
        assert result[0]["s"] == 60

    def test_avg(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 20})
        result = list(db.execute("MATCH (n:N) RETURN avg(n.v) AS a"))
        assert abs(result[0]["a"] - 15.0) < 0.001

    def test_min(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 20})
        result = list(db.execute("MATCH (n:N) RETURN min(n.v) AS m"))
        assert result[0]["m"] == 5

    def test_max(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 20})
        result = list(db.execute("MATCH (n:N) RETURN max(n.v) AS m"))
        assert result[0]["m"] == 20

    def test_collect(self, db):
        """COLLECT aggregates into a list."""
        db.create_node(["N"], {"v": "a"})
        db.create_node(["N"], {"v": "b"})
        result = list(db.execute("MATCH (n:N) RETURN collect(n.v) AS items"))
        assert set(result[0]["items"]) == {"a", "b"}

    def test_stdev(self, db):
        """STDEV() sample standard deviation."""
        for v in [2, 4, 4, 4, 5, 5, 7, 9]:
            db.create_node(["N"], {"v": v})
        result = list(db.execute("MATCH (n:N) RETURN stdev(n.v) AS s"))
        assert abs(result[0]["s"] - 2.0) < 0.2

    def test_stdevp(self, db):
        """STDEVP() population standard deviation."""
        for v in [2, 4, 4, 4, 5, 5, 7, 9]:
            db.create_node(["N"], {"v": v})
        result = list(db.execute("MATCH (n:N) RETURN stdevp(n.v) AS s"))
        assert result[0]["s"] > 0

    def test_percentile_disc(self, db):
        """PERCENTILE_DISC discrete percentile."""
        for v in [1, 2, 3, 4, 5]:
            db.create_node(["N"], {"v": v})
        result = list(db.execute("MATCH (n:N) RETURN percentile_disc(n.v, 0.5) AS p"))
        assert result[0]["p"] == 3

    def test_percentile_cont(self, db):
        """PERCENTILE_CONT continuous percentile."""
        for v in [1, 2, 3, 4, 5]:
            db.create_node(["N"], {"v": v})
        result = list(db.execute("MATCH (n:N) RETURN percentile_cont(n.v, 0.5) AS p"))
        assert abs(result[0]["p"] - 3.0) < 0.1

    def test_listagg(self, db):
        """LISTAGG string aggregation."""
        db.create_node(["N"], {"v": "a"})
        db.create_node(["N"], {"v": "b"})
        db.create_node(["N"], {"v": "c"})
        result = list(db.execute("MATCH (n:N) RETURN listagg(n.v, ', ') AS r"))
        r = result[0]["r"]
        # Order may vary, but should contain all values
        assert "a" in r
        assert "b" in r
        assert "c" in r


# =============================================================================
# String Functions
# =============================================================================


class TestStringFunctions:
    """String manipulation functions."""

    def test_to_upper(self, db):
        db.create_node(["N"], {"v": "hello"})
        result = list(db.execute("MATCH (n:N) RETURN toUpper(n.v) AS r"))
        assert result[0]["r"] == "HELLO"

    def test_upper_alias(self, db):
        """upper() alias for toUpper()."""
        db.create_node(["N"], {"v": "hello"})
        result = list(db.execute("MATCH (n:N) RETURN upper(n.v) AS r"))
        assert result[0]["r"] == "HELLO"

    def test_to_lower(self, db):
        db.create_node(["N"], {"v": "HELLO"})
        result = list(db.execute("MATCH (n:N) RETURN toLower(n.v) AS r"))
        assert result[0]["r"] == "hello"

    def test_lower_alias(self, db):
        db.create_node(["N"], {"v": "HELLO"})
        result = list(db.execute("MATCH (n:N) RETURN lower(n.v) AS r"))
        assert result[0]["r"] == "hello"

    def test_trim(self, db):
        db.create_node(["N"], {"v": "  hello  "})
        result = list(db.execute("MATCH (n:N) RETURN trim(n.v) AS r"))
        assert result[0]["r"] == "hello"

    def test_ltrim(self, db):
        db.create_node(["N"], {"v": "  hello  "})
        result = list(db.execute("MATCH (n:N) RETURN ltrim(n.v) AS r"))
        assert result[0]["r"] == "hello  "

    def test_rtrim(self, db):
        db.create_node(["N"], {"v": "  hello  "})
        result = list(db.execute("MATCH (n:N) RETURN rtrim(n.v) AS r"))
        assert result[0]["r"] == "  hello"

    def test_replace(self, db):
        db.create_node(["N"], {"v": "hello world"})
        result = list(db.execute("MATCH (n:N) RETURN replace(n.v, 'world', 'grafeo') AS r"))
        assert result[0]["r"] == "hello grafeo"

    def test_substring(self, db):
        db.create_node(["N"], {"v": "Amsterdam"})
        result = list(db.execute("MATCH (n:N) RETURN substring(n.v, 0, 5) AS r"))
        assert result[0]["r"] == "Amste"

    def test_split(self, db):
        db.create_node(["N"], {"v": "a,b,c"})
        result = list(db.execute("MATCH (n:N) RETURN split(n.v, ',') AS r"))
        assert result[0]["r"] == ["a", "b", "c"]

    def test_left(self, db):
        db.create_node(["N"], {"v": "Amsterdam"})
        result = list(db.execute("MATCH (n:N) RETURN left(n.v, 3) AS r"))
        assert result[0]["r"] == "Ams"

    def test_right(self, db):
        db.create_node(["N"], {"v": "Amsterdam"})
        result = list(db.execute("MATCH (n:N) RETURN right(n.v, 3) AS r"))
        assert result[0]["r"] == "dam"

    def test_char_length(self, db):
        db.create_node(["N"], {"v": "Amsterdam"})
        result = list(db.execute("MATCH (n:N) RETURN char_length(n.v) AS r"))
        assert result[0]["r"] == 9

    def test_reverse_string(self, db):
        db.create_node(["N"], {"v": "abc"})
        result = list(db.execute("MATCH (n:N) RETURN reverse(n.v) AS r"))
        assert result[0]["r"] == "cba"

    def test_to_string(self, db):
        db.create_node(["N"], {"v": 42})
        result = list(db.execute("MATCH (n:N) RETURN toString(n.v) AS r"))
        assert result[0]["r"] == "42"

    def test_string_join(self, db):
        """string_join(list, sep) list-to-string concatenation."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN string_join(['a', 'b', 'c'], '-') AS r"))
        assert result[0]["r"] == "a-b-c"

    def test_octet_length(self, db):
        """octet_length(str) byte length."""
        db.create_node(["N"], {"v": "abc"})
        result = list(db.execute("MATCH (n:N) RETURN octet_length(n.v) AS r"))
        assert result[0]["r"] == 3


# =============================================================================
# Numeric Functions
# =============================================================================


class TestNumericFunctions:
    """Numeric math functions."""

    def test_abs(self, db):
        db.create_node(["N"], {"v": -5})
        result = list(db.execute("MATCH (n:N) RETURN abs(n.v) AS r"))
        assert result[0]["r"] == 5

    def test_ceil(self, db):
        db.create_node(["N"], {"v": 3.2})
        result = list(db.execute("MATCH (n:N) RETURN ceil(n.v) AS r"))
        assert result[0]["r"] == 4

    def test_ceiling_alias(self, db):
        db.create_node(["N"], {"v": 3.2})
        result = list(db.execute("MATCH (n:N) RETURN ceiling(n.v) AS r"))
        assert result[0]["r"] == 4

    def test_floor(self, db):
        db.create_node(["N"], {"v": 3.8})
        result = list(db.execute("MATCH (n:N) RETURN floor(n.v) AS r"))
        assert result[0]["r"] == 3

    def test_round(self, db):
        db.create_node(["N"], {"v": 3.5})
        result = list(db.execute("MATCH (n:N) RETURN round(n.v) AS r"))
        assert result[0]["r"] == 4

    def test_sqrt(self, db):
        db.create_node(["N"], {"v": 16})
        result = list(db.execute("MATCH (n:N) RETURN sqrt(n.v) AS r"))
        assert result[0]["r"] == 4.0

    def test_sign(self, db):
        db.create_node(["N"], {"v": -5})
        result = list(db.execute("MATCH (n:N) RETURN sign(n.v) AS r"))
        assert result[0]["r"] == -1

    def test_log_ln(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN log(e()) AS r"))
        assert abs(result[0]["r"] - 1.0) < 0.001

    def test_log10(self, db):
        db.create_node(["N"], {"v": 100})
        result = list(db.execute("MATCH (n:N) RETURN log10(n.v) AS r"))
        assert abs(result[0]["r"] - 2.0) < 0.001

    def test_exp(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN exp(n.v) AS r"))
        assert abs(result[0]["r"] - math.e) < 0.001

    def test_rand(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN rand() AS r"))
        assert 0 <= result[0]["r"] <= 1

    def test_sin(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN sin(0) AS r"))
        assert abs(result[0]["r"]) < 0.001

    def test_cos(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN cos(0) AS r"))
        assert abs(result[0]["r"] - 1.0) < 0.001

    def test_tan(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN tan(0) AS r"))
        assert abs(result[0]["r"]) < 0.001

    def test_asin(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN asin(1) AS r"))
        assert abs(result[0]["r"] - math.pi / 2) < 0.001

    def test_acos(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN acos(1) AS r"))
        assert abs(result[0]["r"]) < 0.001

    def test_atan(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN atan(0) AS r"))
        assert abs(result[0]["r"]) < 0.001

    def test_degrees(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN degrees(pi()) AS r"))
        assert abs(result[0]["r"] - 180.0) < 0.001

    def test_radians(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN radians(180) AS r"))
        assert abs(result[0]["r"] - math.pi) < 0.001

    def test_pi(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN pi() AS r"))
        assert abs(result[0]["r"] - math.pi) < 0.001

    def test_e(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN e() AS r"))
        assert abs(result[0]["r"] - math.e) < 0.001


# =============================================================================
# Temporal Functions
# =============================================================================


class TestTemporalFunctions:
    """Temporal construction and extraction."""

    def test_date_no_arg(self, db):
        """date() returns today."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN date() AS r"))
        assert result[0]["r"] is not None

    def test_date_parse(self, db):
        """date(string) parses ISO date."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN date('2024-06-15') AS r"))
        assert "2024-06-15" in str(result[0]["r"])

    def test_time_no_arg(self, db):
        """time() returns current time."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN time() AS r"))
        assert result[0]["r"] is not None

    def test_time_parse(self, db):
        """time(string) parses ISO time."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN time('14:30:00') AS r"))
        assert "14:30" in str(result[0]["r"])

    def test_datetime_parse(self, db):
        """datetime(string) parses ISO datetime."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN datetime('2024-06-15T14:30:00') AS r"))
        assert "2024" in str(result[0]["r"])

    def test_duration_parse(self, db):
        """duration(string) parses ISO duration."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN duration('P1Y2M3D') AS r"))
        assert result[0]["r"] is not None

    def test_current_date(self, db):
        """current_date() alias."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN current_date() AS r"))
        assert result[0]["r"] is not None

    def test_current_time(self, db):
        """current_time() alias."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN current_time() AS r"))
        assert result[0]["r"] is not None

    def test_now(self, db):
        """now() returns current timestamp."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN now() AS r"))
        assert result[0]["r"] is not None

    def test_timestamp_returns_millis(self, db):
        """timestamp() returns millisecond epoch as integer (#179)."""
        result = list(db.execute("RETURN timestamp() AS ts"))
        assert len(result) == 1
        assert result[0]["ts"] is not None
        assert isinstance(result[0]["ts"], int)
        assert result[0]["ts"] > 1_577_836_800_000  # after 2020-01-01

    def test_timestamp_in_on_create_set(self, db):
        """MERGE ON CREATE SET n.ts = timestamp() stores non-null value (#179)."""
        db.execute("MERGE (n:T {id: 'x'}) ON CREATE SET n.ts = timestamp()")
        result = list(db.execute("MATCH (n:T) RETURN n.ts"))
        assert result[0]["n.ts"] is not None
        assert result[0]["n.ts"] > 1_577_836_800_000

    def test_year_extraction(self, db):
        """year(date) extracts year."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN year(date('2024-06-15')) AS r"))
        assert result[0]["r"] == 2024

    def test_month_extraction(self, db):
        """month(date) extracts month."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN month(date('2024-06-15')) AS r"))
        assert result[0]["r"] == 6

    def test_day_extraction(self, db):
        """day(date) extracts day."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN day(date('2024-06-15')) AS r"))
        assert result[0]["r"] == 15

    def test_hour_extraction(self, db):
        """hour(time) extracts hour."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN hour(time('14:30:00')) AS r"))
        assert result[0]["r"] == 14

    def test_minute_extraction(self, db):
        """minute(time) extracts minute."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN minute(time('14:30:00')) AS r"))
        assert result[0]["r"] == 30

    def test_second_extraction(self, db):
        """second(time) extracts second."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN second(time('14:30:45')) AS r"))
        assert result[0]["r"] == 45

    def test_temporal_arithmetic(self, db):
        """date + duration temporal arithmetic."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN date('2024-01-15') + duration('P10D') AS r"))
        assert "2024-01-25" in str(result[0]["r"])

    def test_local_time_alias(self, db):
        """local_time() is alias for time()."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN local_time() AS r"))
        assert result[0]["r"] is not None

    def test_local_datetime_alias(self, db):
        """local_datetime() is alias for datetime()."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN local_datetime() AS r"))
        assert result[0]["r"] is not None

    def test_zoned_datetime_no_arg(self, db):
        """zoned_datetime() returns current UTC."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN zoned_datetime() AS r"))
        assert result[0]["r"] is not None

    def test_date_trunc(self, db):
        """date_trunc(unit, temporal) truncation."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN date_trunc('month', date('2024-06-15')) AS r"))
        assert "2024-06-01" in str(result[0]["r"])


# =============================================================================
# List Functions
# =============================================================================


class TestListFunctions:
    """List manipulation functions."""

    def test_head(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN head([1, 2, 3]) AS r"))
        assert result[0]["r"] == 1

    def test_tail(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN tail([1, 2, 3]) AS r"))
        assert result[0]["r"] == [2, 3]

    def test_last(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN last([1, 2, 3]) AS r"))
        assert result[0]["r"] == 3

    def test_reverse_list(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN reverse([1, 2, 3]) AS r"))
        assert result[0]["r"] == [3, 2, 1]

    def test_size(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN size([1, 2, 3]) AS r"))
        assert result[0]["r"] == 3

    def test_range_two_args(self, db):
        """range(start, end)."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN range(1, 5) AS r"))
        assert result[0]["r"] == [1, 2, 3, 4, 5]

    def test_range_with_step(self, db):
        """range(start, end, step)."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN range(0, 10, 3) AS r"))
        assert result[0]["r"] == [0, 3, 6, 9]

    def test_all_predicate(self, db):
        """all(x IN list WHERE pred) list predicate."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN all(x IN [2, 4, 6] WHERE x % 2 = 0) AS r"))
        assert result[0]["r"] is True

    def test_any_predicate(self, db):
        """any(x IN list WHERE pred) list predicate."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN any(x IN [1, 2, 3] WHERE x > 2) AS r"))
        assert result[0]["r"] is True

    def test_none_predicate(self, db):
        """none(x IN list WHERE pred) list predicate."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN none(x IN [1, 2, 3] WHERE x > 10) AS r"))
        assert result[0]["r"] is True

    def test_single_predicate(self, db):
        """single(x IN list WHERE pred) exactly one match."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute("MATCH (n:N) RETURN single(x IN [1, 2, 3] WHERE x = 2) AS r"))
        assert result[0]["r"] is True


# =============================================================================
# Graph Element Functions
# =============================================================================


class TestGraphElementFunctions:
    """id, element_id, labels, type, keys, properties, hasLabel."""

    def test_id(self, db):
        db.create_node(["Person"], {"name": "Alix"})
        result = list(db.execute("MATCH (n:Person) RETURN id(n) AS r"))
        assert isinstance(result[0]["r"], int)

    def test_element_id(self, db):
        db.create_node(["Person"], {"name": "Alix"})
        result = list(db.execute("MATCH (n:Person) RETURN element_id(n) AS r"))
        assert result[0]["r"] is not None

    def test_labels(self, db):
        db.create_node(["Person", "Developer"], {"name": "Alix"})
        result = list(db.execute("MATCH (n:Person) RETURN labels(n) AS r"))
        assert "Person" in result[0]["r"]
        assert "Developer" in result[0]["r"]

    def test_type_edge(self, db):
        a = db.create_node(["Person"], {"name": "Alix"})
        b = db.create_node(["Person"], {"name": "Gus"})
        db.create_edge(a.id, b.id, "KNOWS")
        result = list(db.execute("MATCH ()-[e:KNOWS]->() RETURN type(e) AS r"))
        assert result[0]["r"] == "KNOWS"

    def test_keys(self, db):
        db.create_node(["Person"], {"name": "Alix", "age": 30})
        result = list(db.execute("MATCH (n:Person) RETURN keys(n) AS r"))
        assert "name" in result[0]["r"]
        assert "age" in result[0]["r"]

    def test_properties(self, db):
        db.create_node(["Person"], {"name": "Alix", "age": 30})
        result = list(db.execute("MATCH (n:Person) RETURN properties(n) AS r"))
        assert result[0]["r"]["name"] == "Alix"
        assert result[0]["r"]["age"] == 30

    def test_has_label(self, db):
        db.create_node(["Person"], {"name": "Alix"})
        result = list(db.execute("MATCH (n:Person) RETURN hasLabel(n, 'Person') AS r"))
        assert result[0]["r"] is True

    def test_start_node(self, db):
        """startNode(r) returns source node ID (#180)."""
        db.execute("INSERT (:Person {name: 'Alix'})-[:KNOWS]->(:Person {name: 'Gus'})")
        result = list(db.execute("MATCH ()-[r:KNOWS]->() RETURN startNode(r) AS sn"))
        assert len(result) == 1
        assert result[0]["sn"] is not None
        assert isinstance(result[0]["sn"], int)

    def test_end_node(self, db):
        """endNode(r) returns destination node ID (#180)."""
        db.execute("INSERT (:Person {name: 'Alix'})-[:KNOWS]->(:Person {name: 'Gus'})")
        result = list(db.execute("MATCH ()-[r:KNOWS]->() RETURN endNode(r) AS en"))
        assert len(result) == 1
        assert result[0]["en"] is not None
        assert isinstance(result[0]["en"], int)

    def test_start_node_matches_source(self, db):
        """startNode(r) == id(s) and endNode(r) == id(t) (#180)."""
        db.execute("INSERT (:Person {name: 'Alix'})-[:KNOWS]->(:Person {name: 'Gus'})")
        result = list(
            db.execute(
                "MATCH (s)-[r:KNOWS]->(t) "
                "RETURN startNode(r) = id(s) AS src_ok, endNode(r) = id(t) AS dst_ok"
            )
        )
        assert result[0]["src_ok"] is True
        assert result[0]["dst_ok"] is True


# =============================================================================
# Type Conversion Functions
# =============================================================================


class TestTypeConversion:
    """toInteger, toFloat, toString, toBoolean, toDate, toTime, etc."""

    def test_to_integer(self, db):
        db.create_node(["N"], {"v": "42"})
        result = list(db.execute("MATCH (n:N) RETURN toInteger(n.v) AS r"))
        assert result[0]["r"] == 42

    def test_to_int_alias(self, db):
        db.create_node(["N"], {"v": "42"})
        result = list(db.execute("MATCH (n:N) RETURN toInt(n.v) AS r"))
        assert result[0]["r"] == 42

    def test_to_float(self, db):
        db.create_node(["N"], {"v": "3.14"})
        result = list(db.execute("MATCH (n:N) RETURN toFloat(n.v) AS r"))
        assert abs(result[0]["r"] - 3.14) < 0.001

    def test_to_boolean(self, db):
        db.create_node(["N"], {"v": "true"})
        result = list(db.execute("MATCH (n:N) RETURN toBoolean(n.v) AS r"))
        assert result[0]["r"] is True

    def test_to_date(self, db):
        db.create_node(["N"], {"v": "2024-06-15"})
        result = list(db.execute("MATCH (n:N) RETURN toDate(n.v) AS r"))
        assert "2024-06-15" in str(result[0]["r"])

    def test_to_time(self, db):
        db.create_node(["N"], {"v": "14:30:00"})
        result = list(db.execute("MATCH (n:N) RETURN toTime(n.v) AS r"))
        assert "14:30" in str(result[0]["r"])

    def test_to_datetime(self, db):
        db.create_node(["N"], {"v": "2024-06-15T14:30:00"})
        result = list(db.execute("MATCH (n:N) RETURN toDatetime(n.v) AS r"))
        assert "2024" in str(result[0]["r"])

    def test_to_duration(self, db):
        db.create_node(["N"], {"v": "P1Y2M"})
        result = list(db.execute("MATCH (n:N) RETURN toDuration(n.v) AS r"))
        assert result[0]["r"] is not None

    def test_to_list(self, db):
        """toList wraps scalar in list."""
        db.create_node(["N"], {"v": 42})
        result = list(db.execute("MATCH (n:N) RETURN toList(n.v) AS r"))
        assert result[0]["r"] == [42]



"""Cypher implementation of solvOR plugin comparison tests.

Compares Grafeo's as_solvor() plugin against standalone solvOR library.
Uses Python API for graph setup (Cypher execution not required).
"""

import random

import pytest
from grafeo import GrafeoDB

from tests.bases.test_solvor import (
    BaseSolvORBenchmarkTest,
    BaseSolvORComparisonTest,
)


@pytest.fixture
def db():
    """Create a fresh database instance."""
    return GrafeoDB()


class TestCypherSolvORComparison(BaseSolvORComparisonTest):
    """Cypher implementation of solvOR plugin comparison tests.

    Tests Grafeo's as_solvor() plugin against standalone solvOR library.
    Graph setup uses Python API (Cypher execution not required).
    """

    def create_db(self):
        """Create a fresh database instance."""
        return GrafeoDB()

    def setup_flow_network(self, db, n_nodes: int, n_edges: int, seed: int = 42) -> dict:
        """Set up a flow network using Python API."""
        rng = random.Random(seed)

        node_ids = []
        for i in range(n_nodes):
            node = db.create_node(["Node"], {"index": i})
            node_ids.append(node.id)

        source = node_ids[0]
        sink = node_ids[-1]

        edges: list[tuple] = []
        edge_set: set[tuple] = set()

        # Create backbone path to ensure connectivity
        for i in range(min(5, n_nodes - 1)):
            src = node_ids[i]
            dst = node_ids[i + 1]
            if (src, dst) not in edge_set:
                capacity = rng.randint(10, 50)
                cost = rng.randint(1, 10)
                db.create_edge(src, dst, "FLOW", {"capacity": capacity, "cost": cost})
                edges.append((src, dst, capacity, cost))
                edge_set.add((src, dst))

        # Add remaining random edges
        attempts = 0
        while len(edges) < n_edges and attempts < n_edges * 3:
            attempts += 1
            src = rng.choice(node_ids)
            dst = rng.choice(node_ids)
            if src != dst and (src, dst) not in edge_set and (dst, src) not in edge_set:
                capacity = rng.randint(5, 30)
                cost = rng.randint(1, 15)
                db.create_edge(src, dst, "FLOW", {"capacity": capacity, "cost": cost})
                edges.append((src, dst, capacity, cost))
                edge_set.add((src, dst))

        return {"node_ids": node_ids, "source": source, "sink": sink, "edges": edges}


class TestCypherSolvORBenchmark(BaseSolvORBenchmarkTest):
    """Cypher implementation of solvOR plugin benchmark tests."""

    def create_db(self):
        """Create a fresh database instance."""
        return GrafeoDB()

    def setup_flow_network(self, db, n_nodes: int, n_edges: int, seed: int = 42) -> dict:
        """Set up a flow network using Python API."""
        rng = random.Random(seed)

        node_ids = []
        for i in range(n_nodes):
            node = db.create_node(["Node"], {"index": i})
            node_ids.append(node.id)

        source = node_ids[0]
        sink = node_ids[-1]

        edges: list[tuple] = []
        edge_set: set[tuple] = set()

        # Create backbone path
        for i in range(min(5, n_nodes - 1)):
            src = node_ids[i]
            dst = node_ids[i + 1]
            if (src, dst) not in edge_set:
                capacity = rng.randint(10, 50)
                cost = rng.randint(1, 10)
                db.create_edge(src, dst, "FLOW", {"capacity": capacity, "cost": cost})
                edges.append((src, dst, capacity, cost))
                edge_set.add((src, dst))

        # Add remaining random edges
        attempts = 0
        while len(edges) < n_edges and attempts < n_edges * 3:
            attempts += 1
            src = rng.choice(node_ids)
            dst = rng.choice(node_ids)
            if src != dst and (src, dst) not in edge_set and (dst, src) not in edge_set:
                capacity = rng.randint(5, 30)
                cost = rng.randint(1, 15)
                db.create_edge(src, dst, "FLOW", {"capacity": capacity, "cost": cost})
                edges.append((src, dst, capacity, cost))
                edge_set.add((src, dst))

        return {"node_ids": node_ids, "source": source, "sink": sink, "edges": edges}



"""Cypher implementation of algorithm benchmarks.

Benchmarks graph algorithms using db.algorithms.* API.
Cypher is used for setup only.
"""

import random

from tests.bases.bench_algorithms import BaseBenchAlgorithms


class BenchCypherAlgorithms(BaseBenchAlgorithms):
    """Cypher implementation of algorithm benchmarks.

    Note: Algorithms are accessed via db.algorithms.*, not via Cypher queries.
    Cypher is used for graph setup only.
    """

    def setup_random_graph(self, db, n_nodes: int, n_edges: int, weighted: bool = True) -> dict:
        """Set up a random graph for benchmarking."""
        rng = random.Random(42)

        node_ids = []
        for i in range(n_nodes):
            node = db.create_node(["Node"], {"index": i})
            node_ids.append(node.id)

        edges = set()
        while len(edges) < n_edges:
            src = rng.choice(node_ids)
            dst = rng.choice(node_ids)
            if src != dst and (src, dst) not in edges:
                props = {"weight": rng.uniform(0.1, 10.0)} if weighted else {}
                db.create_edge(src, dst, "EDGE", props)
                edges.add((src, dst))

        return {"node_ids": node_ids, "edge_count": len(edges)}

    def run_bfs(self, db, start_node) -> list:
        """Run BFS from start node."""
        return db.algorithms.bfs(start_node)

    def run_dfs(self, db, start_node) -> list:
        """Run DFS from start node."""
        return db.algorithms.dfs(start_node)

    def run_dijkstra(self, db, source, target=None, weight_prop: str = "weight"):
        """Run Dijkstra's algorithm."""
        if target is None:
            return db.algorithms.dijkstra(source, weight=weight_prop)
        return db.algorithms.dijkstra(source, target, weight_prop)

    def run_bellman_ford(self, db, source, weight_prop: str = "weight"):
        """Run Bellman-Ford algorithm."""
        return db.algorithms.bellman_ford(source, weight_prop)

    def run_connected_components(self, db) -> dict:
        """Run connected components."""
        return db.algorithms.connected_components()

    def run_strongly_connected_components(self, db) -> list:
        """Run strongly connected components."""
        return db.algorithms.strongly_connected_components()

    def run_pagerank(self, db, damping: float = 0.85, iterations: int = 20) -> dict:
        """Run PageRank algorithm."""
        return db.algorithms.pagerank(damping=damping, max_iterations=iterations)

    def run_degree_centrality(self, db, normalized: bool = True) -> dict:
        """Run degree centrality."""
        return db.algorithms.degree_centrality(normalized=normalized)

    def run_betweenness_centrality(self, db) -> dict:
        """Run betweenness centrality."""
        return db.algorithms.betweenness_centrality()

    def run_closeness_centrality(self, db) -> dict:
        """Run closeness centrality."""
        return db.algorithms.closeness_centrality()

    def run_label_propagation(self, db) -> dict:
        """Run label propagation community detection."""
        return db.algorithms.label_propagation()

    def run_louvain(self, db) -> dict:
        """Run Louvain community detection."""
        return db.algorithms.louvain()

    def run_kruskal(self, db, weight_prop: str = "weight") -> dict:
        """Run Kruskal's MST algorithm."""
        return db.algorithms.kruskal(weight_prop)

    def run_prim(self, db, weight_prop: str = "weight") -> dict:
        """Run Prim's MST algorithm."""
        return db.algorithms.prim(weight_prop)



"""Abstract base test classes for Grafeo test suite.

This module contains base classes that define WHAT to test. Each language
implementation inherits from these and provides HOW (the query syntax).

Base Test Classes:
- BaseMutationsTest: CRUD operations
- BaseTransactionsTest: Transaction handling
- BaseAlgorithmsTest: Graph algorithms

Base Benchmark Classes:
- BaseBenchStorage: Storage benchmarks (reads + writes)
- BaseBenchAlgorithms: Algorithm benchmarks

Comparison Test Classes:
- BaseNetworkXComparisonTest: Compare against NetworkX reference
- BaseNetworkXBenchmarkTest: Benchmark against NetworkX
- BaseSolvORComparisonTest: Compare against OR-Tools reference
- BaseSolvORBenchmarkTest: Benchmark against OR-Tools
"""

from .bench_algorithms import BaseBenchAlgorithms
from .bench_storage import BaseBenchStorage, BenchmarkResult
from .test_algorithms import BaseAlgorithmsTest
from .test_mutations import BaseMutationsTest
from .test_networkx import BaseNetworkXBenchmarkTest, BaseNetworkXComparisonTest
from .test_solvor import BaseSolvORBenchmarkTest, BaseSolvORComparisonTest
from .test_transactions import BaseTransactionsTest

__all__ = [
    # Test base classes
    "BaseMutationsTest",
    "BaseTransactionsTest",
    "BaseAlgorithmsTest",
    # Benchmark base classes
    "BaseBenchStorage",
    "BaseBenchAlgorithms",
    "BenchmarkResult",
    # Comparison test classes
    "BaseNetworkXComparisonTest",
    "BaseNetworkXBenchmarkTest",
    "BaseSolvORComparisonTest",
    "BaseSolvORBenchmarkTest",
]



"""Base class for mutation tests.

This module defines test logic for all write operations:
- Create nodes (single, multiple labels, with properties)
- Create edges
- Update nodes/edges
- Delete nodes/edges
"""

from abc import ABC, abstractmethod


class BaseMutationsTest(ABC):
    """Abstract base class for mutation tests.

    Subclasses implement query builders for their specific language.
    """

    # =========================================================================
    # EXECUTION
    # =========================================================================

    def execute_query(self, db, query):
        """Execute a query using the appropriate language parser.

        Override in subclasses that need a specific parser (e.g., Cypher).
        Default uses GQL parser via db.execute().
        """
        return db.execute(query)

    # =========================================================================
    # QUERY BUILDERS
    # =========================================================================

    @abstractmethod
    def create_node_query(self, labels: list[str], props: dict) -> str:
        """Return query to create a node with given labels and properties.

        Args:
            labels: List of node labels (e.g., ["Person", "Employee"])
            props: Dictionary of properties (e.g., {"name": "Alix", "age": 30})

        Returns:
            Language-specific query string
        """
        raise NotImplementedError

    @abstractmethod
    def match_node_query(self, label: str, return_prop: str = "name") -> str:
        """Return query to match nodes by label and return a property.

        Args:
            label: Node label to match
            return_prop: Property to return

        Returns:
            Language-specific query string
        """
        raise NotImplementedError

    @abstractmethod
    def match_where_query(
        self, label: str, prop: str, op: str, value, return_prop: str = "name"
    ) -> str:
        """Return query to match nodes with WHERE clause.

        Args:
            label: Node label to match
            prop: Property to filter on
            op: Comparison operator (e.g., ">", "=", "<")
            value: Value to compare against
            return_prop: Property to return

        Returns:
            Language-specific query string
        """
        raise NotImplementedError

    @abstractmethod
    def delete_node_query(self, label: str, prop: str, value) -> str:
        """Return query to delete a node matching criteria.

        Args:
            label: Node label to match
            prop: Property to match on
            value: Property value to match

        Returns:
            Language-specific query string
        """
        raise NotImplementedError

    @abstractmethod
    def create_edge_query(
        self,
        from_label: str,
        from_prop: str,
        from_value,
        to_label: str,
        to_prop: str,
        to_value,
        edge_type: str,
        edge_props: dict,
    ) -> str:
        """Return query to create an edge between two nodes.

        Args:
            from_label: Source node label
            from_prop: Source node property to match
            from_value: Source node property value
            to_label: Target node label
            to_prop: Target node property to match
            to_value: Target node property value
            edge_type: Edge/relationship type
            edge_props: Edge properties

        Returns:
            Language-specific query string
        """
        raise NotImplementedError

    @abstractmethod
    def update_node_query(
        self, label: str, match_prop: str, match_value, set_prop: str, set_value
    ) -> str:
        """Return query to update a node property.

        Args:
            label: Node label to match
            match_prop: Property to match on
            match_value: Value to match
            set_prop: Property to set
            set_value: New value

        Returns:
            Language-specific query string
        """
        raise NotImplementedError

    # =========================================================================
    # CREATE TESTS
    # =========================================================================

    def test_create_single_node(self, db):
        """Test creating a single node."""
        query = self.create_node_query(["Person"], {"name": "Alix", "age": 30})
        result = self.execute_query(db, query)
        rows = list(result)
        assert len(rows) >= 0  # INSERT may not return rows

        # Verify node exists
        match_query = self.match_node_query("Person")
        result = self.execute_query(db, match_query)
        rows = list(result)
        assert len(rows) == 1

    def test_create_node_multiple_labels(self, db):
        """Test creating a node with multiple labels."""
        query = self.create_node_query(["Person", "Developer"], {"name": "Gus", "language": "Rust"})
        self.execute_query(db, query)

        # Verify node has both labels
        match_query = self.match_node_query("Person")
        result = self.execute_query(db, match_query)
        rows = list(result)
        names = [r.get("n.name") or r.get("name") for r in rows]
        assert "Gus" in names

    def test_create_node_with_properties(self, db):
        """Test creating a node with various property types."""
        query = self.create_node_query(
            ["Data"],
            {
                "string_val": "hello",
                "int_val": 42,
                "float_val": 3.14,
                "bool_val": True,
            },
        )
        self.execute_query(db, query)

        # Verify properties exist
        result = self.execute_query(db, "MATCH (n:Data) RETURN n.int_val")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0].get("n.int_val") == 42

    def test_create_multiple_nodes(self, db):
        """Test creating multiple nodes."""
        for i in range(5):
            query = self.create_node_query(["Person"], {"name": f"Person{i}", "idx": i})
            self.execute_query(db, query)

        # Verify all nodes created
        match_query = self.match_node_query("Person")
        result = self.execute_query(db, match_query)
        rows = list(result)
        assert len(rows) == 5

    def test_create_edge(self, db):
        """Test creating an edge between nodes."""
        # Create two nodes
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Alix"}))
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Gus"}))

        # Create edge
        query = self.create_edge_query(
            "Person", "name", "Alix", "Person", "name", "Gus", "KNOWS", {"since": 2020}
        )
        self.execute_query(db, query)

        # Verify edge exists
        result = self.execute_query(
            db, "MATCH (a:Person)-[r:KNOWS]->(b:Person) RETURN a.name, b.name, r.since"
        )
        rows = list(result)
        assert len(rows) == 1
        assert rows[0].get("r.since") == 2020

    # =========================================================================
    # UPDATE TESTS
    # =========================================================================

    def test_update_node_property(self, db):
        """Test updating a node property."""
        # Create node
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Alix", "age": 30}))

        # Update age
        query = self.update_node_query("Person", "name", "Alix", "age", 31)
        self.execute_query(db, query)

        # Verify update
        result = self.execute_query(db, "MATCH (n:Person {name: 'Alix'}) RETURN n.age")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0].get("n.age") == 31

    # =========================================================================
    # DELETE TESTS
    # =========================================================================

    def test_delete_node(self, db):
        """Test deleting a node."""
        # Create nodes
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Alix"}))
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Gus"}))

        # Delete Alix
        query = self.delete_node_query("Person", "name", "Alix")
        self.execute_query(db, query)

        # Verify only Gus remains
        match_query = self.match_node_query("Person")
        result = self.execute_query(db, match_query)
        rows = list(result)
        names = [r.get("n.name") or r.get("name") for r in rows]
        assert len(rows) == 1
        assert "Gus" in names
        assert "Alix" not in names

    # =========================================================================
    # FILTER TESTS (part of mutations workflow)
    # =========================================================================

    def test_match_with_filter(self, db):
        """Test matching nodes with WHERE clause."""
        # Create test data
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Alix", "age": 30}))
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Gus", "age": 25}))
        self.execute_query(db, self.create_node_query(["Person"], {"name": "Vincent", "age": 35}))

        # Query with filter
        query = self.match_where_query("Person", "age", ">", 28)
        result = self.execute_query(db, query)
        rows = list(result)

        # Alix (30) and Vincent (35) should match
        assert len(rows) == 2



"""GraphQL implementation of transaction tests.

Tests transaction commit, rollback, and isolation using GraphQL syntax.
"""

import pytest

from tests.bases.test_transactions import BaseTransactionsTest


def execute_graphql(db, query: str):
    """Execute GraphQL query, skip if not supported."""
    try:
        return db.execute_graphql(query)
    except AttributeError:
        pytest.skip("GraphQL support not available in this build")
        return None
    except NotImplementedError:
        pytest.skip("GraphQL not implemented")
        return None


class TestGraphQLTransactions(BaseTransactionsTest):
    """GraphQL implementation of transaction tests.

    Note: Uses begin_transaction() with GraphQL queries.
    """

    def insert_query(self, labels: list[str], props: dict) -> str:
        """Return GraphQL create mutation."""
        label = labels[0] if labels else "Node"
        args = ", ".join(
            f'{k}: "{v}"' if isinstance(v, str) else f"{k}: {v}" for k, v in props.items()
        )
        return f"""
            mutation {{
                create{label}({args}) {{
                    id
                }}
            }}
        """

    def match_by_prop_query(self, label: str, prop: str, value) -> str:
        """Return GraphQL query."""
        val = f'"{value}"' if isinstance(value, str) else value
        return f"""
            query {{
                {label.lower()}({prop}: {val}) {{
                    id
                }}
            }}
        """

    def count_query(self, label: str) -> str:
        """Return GraphQL query that returns all rows (base class will count them)."""
        return f"""
            query {{
                {label.lower()} {{
                    id
                }}
            }}
        """

    def execute_in_tx(self, tx, query: str):
        """Execute query in transaction context."""
        try:
            return tx.execute_graphql(query)
        except AttributeError:
            pytest.skip("GraphQL support not available")
            return None

    def execute_query(self, db, query: str):
        """Execute query on database."""
        return execute_graphql(db, query)



"""Cypher-specific regression tests.

Covers correlated EXISTS subqueries, CASE WHEN inside aggregates,
any() with IN lists, CASE WHEN inside reduce(), and target node
property filters on edge patterns.
"""


# =============================================================================
# Bug 1: Correlated EXISTS subquery
# =============================================================================


class TestCorrelatedExistsSubquery:
    """Correlated EXISTS subqueries where inner pattern references outer MATCH variables."""

    def test_not_exists_with_type_filter(self, db):
        """NOT EXISTS with WHERE type(r) filter should exclude matched pairs."""
        db.execute_cypher("CREATE (a {name: 'Alix'}), (b {name: 'Gus'}), (c {name: 'Vincent'})")
        db.execute_cypher("MATCH (a {name: 'Alix'}), (b {name: 'Gus'}) CREATE (a)-[:KNOWS]->(b)")
        db.execute_cypher(
            "MATCH (a {name: 'Alix'}), (c {name: 'Vincent'}) CREATE (a)-[:WORKS_WITH]->(c)"
        )

        result = list(
            db.execute_cypher(
                "MATCH (x), (y) WHERE x <> y "
                "AND NOT EXISTS { MATCH (x)-[r]->(y) WHERE type(r) = 'KNOWS' } "
                "RETURN x.name, y.name"
            )
        )
        pairs = {(r["x.name"], r["y.name"]) for r in result}
        # (Alix, Gus) should be excluded (KNOWS edge exists)
        assert ("Alix", "Gus") not in pairs
        # (Alix, Vincent) should be included (edge is WORKS_WITH, not KNOWS)
        assert ("Alix", "Vincent") in pairs

    def test_not_exists_bare_pattern(self, db):
        """NOT EXISTS with bare pattern (no MATCH keyword) should work identically."""
        db.execute_cypher("CREATE (a {name: 'Alix'}), (b {name: 'Gus'}), (c {name: 'Vincent'})")
        db.execute_cypher("MATCH (a {name: 'Alix'}), (b {name: 'Gus'}) CREATE (a)-[:KNOWS]->(b)")
        db.execute_cypher(
            "MATCH (a {name: 'Alix'}), (c {name: 'Vincent'}) CREATE (a)-[:WORKS_WITH]->(c)"
        )

        result = list(
            db.execute_cypher(
                "MATCH (x), (y) WHERE x <> y "
                "AND NOT EXISTS { (x)-[r]->(y) WHERE type(r) = 'KNOWS' } "
                "RETURN x.name, y.name"
            )
        )
        pairs = {(r["x.name"], r["y.name"]) for r in result}
        assert ("Alix", "Gus") not in pairs
        assert ("Alix", "Vincent") in pairs


# =============================================================================
# Bug 2: CASE WHEN inside aggregate functions
# =============================================================================


class TestCaseWhenInAggregate:
    """CASE WHEN expressions used inside aggregate functions like sum()."""

    def test_sum_case_when(self, db):
        """sum(CASE WHEN ... THEN 1 ELSE 0 END) should count matching rows."""
        db.execute_cypher("CREATE (d:Dir {name: 'src'})")
        db.execute_cypher("CREATE (f1:File {name: 'a.py', file_type: 'source'})")
        db.execute_cypher("CREATE (f2:File {name: 'b.md', file_type: 'docs'})")
        db.execute_cypher("CREATE (f3:File {name: 'c.py', file_type: 'source'})")
        db.execute_cypher(
            "MATCH (d:Dir {name: 'src'}), (f:File {name: 'a.py'}) CREATE (d)-[:CONTAINS]->(f)"
        )
        db.execute_cypher(
            "MATCH (d:Dir {name: 'src'}), (f:File {name: 'b.md'}) CREATE (d)-[:CONTAINS]->(f)"
        )
        db.execute_cypher(
            "MATCH (d:Dir {name: 'src'}), (f:File {name: 'c.py'}) CREATE (d)-[:CONTAINS]->(f)"
        )

        result = list(
            db.execute_cypher(
                "MATCH (d:Dir)-[:CONTAINS]->(f:File) "
                "WITH d, count(f) AS total, "
                "     sum(CASE WHEN f.file_type = 'source' THEN 1 ELSE 0 END) AS source_count "
                "RETURN d.name AS name, total, source_count"
            )
        )
        assert len(result) == 1
        assert result[0]["name"] == "src"
        assert result[0]["total"] == 3
        assert result[0]["source_count"] == 2


# =============================================================================
# Bug 3: any() with IN list
# =============================================================================


class TestAnyWithInList:
    """any() list predicate with IN list comparison inside WHERE clause."""

    def test_any_labels_in_list_matches(self, db):
        """any(lbl IN labels(n) WHERE lbl IN ['A', 'B']) should match node with label A."""
        db.execute_cypher("CREATE (:A:B:C {name: 'Test'})")

        result = list(
            db.execute_cypher(
                "MATCH (n) WHERE any(lbl IN labels(n) WHERE lbl IN ['A', 'B']) RETURN n.name"
            )
        )
        assert len(result) == 1
        assert result[0]["n.name"] == "Test"

    def test_any_labels_in_list_no_match(self, db):
        """any() should return 0 rows when no labels match the IN list."""
        db.execute_cypher("CREATE (:A:B:C {name: 'Test'})")

        result = list(
            db.execute_cypher(
                "MATCH (n) WHERE any(lbl IN labels(n) WHERE lbl IN ['X', 'Y']) RETURN n.name"
            )
        )
        assert len(result) == 0

    def test_any_with_single_match(self, db):
        """any() should work when only one label matches from the IN list."""
        db.execute_cypher("CREATE (:Model:Component {name: 'Widget'})")

        result = list(
            db.execute_cypher(
                "MATCH (n) WHERE any(lbl IN labels(n) WHERE lbl IN ['Model', 'Service']) "
                "RETURN n.name"
            )
        )
        assert len(result) == 1
        assert result[0]["n.name"] == "Widget"


# =============================================================================
# Bug 4: CASE WHEN inside reduce()
# =============================================================================


class TestCaseWhenInReduce:
    """CASE WHEN expressions used inside reduce() accumulator body."""

    def test_reduce_with_case_max(self, db):
        """reduce() with CASE WHEN should compute max value."""
        result = list(
            db.execute_cypher(
                "WITH [3, 1, 4, 1, 5] AS vals "
                "RETURN reduce(acc = 0, x IN vals | "
                "  CASE WHEN x > acc THEN x ELSE acc END) AS max_val"
            )
        )
        assert len(result) == 1
        assert result[0]["max_val"] == 5

    def test_reduce_with_case_min(self, db):
        """reduce() with CASE WHEN should compute min value."""
        result = list(
            db.execute_cypher(
                "WITH [3, 1, 4, 1, 5] AS vals "
                "RETURN reduce(acc = 999, x IN vals | "
                "  CASE WHEN x < acc THEN x ELSE acc END) AS min_val"
            )
        )
        assert len(result) == 1
        assert result[0]["min_val"] == 1

    def test_reduce_with_case_conditional_sum(self, db):
        """reduce() with CASE WHEN should conditionally sum values."""
        result = list(
            db.execute_cypher(
                "WITH [1, 2, 3, 4, 5] AS vals "
                "RETURN reduce(acc = 0, x IN vals | "
                "  CASE WHEN x > 2 THEN acc + x ELSE acc END) AS big_sum"
            )
        )
        assert len(result) == 1
        # Only 3+4+5 = 12
        assert result[0]["big_sum"] == 12


# =============================================================================
# Bug 5: Target node property filter ignored in edge patterns (#155)
# =============================================================================


class TestTargetNodePropertyFilter:
    """Target node property filters like ()-[r]->(o {name: 'X'}) must filter results."""

    def test_outgoing_target_property_filter(self, db):
        """MATCH (a)-[r]->(b {name: 'Gus'}) should only return edges to Gus."""
        db.execute_cypher(
            "CREATE (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}), "
            "(c:Person {name: 'Vincent'})"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}) CREATE (a)-[:KNOWS]->(b)"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (c:Person {name: 'Vincent'}) CREATE (a)-[:KNOWS]->(c)"
        )

        result = list(db.execute_cypher("MATCH (a)-[r]->(b {name: 'Gus'}) RETURN a.name, b.name"))
        assert len(result) == 1
        assert result[0]["a.name"] == "Alix"
        assert result[0]["b.name"] == "Gus"

    def test_incoming_target_property_filter(self, db):
        """MATCH (a)-[r]->(b {name: 'Gus'}) from the target's perspective via incoming."""
        db.execute_cypher(
            "CREATE (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}), "
            "(c:Person {name: 'Vincent'})"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}) CREATE (a)-[:KNOWS]->(b)"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (c:Person {name: 'Vincent'}) CREATE (a)-[:KNOWS]->(c)"
        )

        result = list(db.execute_cypher("MATCH (b {name: 'Gus'})<-[r]-(a) RETURN a.name, b.name"))
        assert len(result) == 1
        assert result[0]["a.name"] == "Alix"
        assert result[0]["b.name"] == "Gus"

    def test_target_property_filter_count(self, db):
        """count(r) with target property filter must not return ALL edges."""
        db.execute_cypher(
            "CREATE (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}), "
            "(c:Person {name: 'Vincent'})"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}) CREATE (a)-[:KNOWS]->(b)"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (c:Person {name: 'Vincent'}) CREATE (a)-[:KNOWS]->(c)"
        )

        result = list(db.execute_cypher("MATCH ()-[r]->(o {name: 'Gus'}) RETURN count(r) AS cnt"))
        assert len(result) == 1
        assert result[0]["cnt"] == 1

    def test_target_property_filter_no_match(self, db):
        """Target property filter that matches nothing should return 0 rows."""
        db.execute_cypher("CREATE (:Person {name: 'Alix'})-[:KNOWS]->(:Person {name: 'Gus'})")

        result = list(
            db.execute_cypher("MATCH ()-[r]->(o {name: 'Nobody'}) RETURN count(r) AS cnt")
        )
        assert len(result) == 1
        assert result[0]["cnt"] == 0

    def test_edge_property_filter(self, db):
        """Edge property filter -[r {since: 2020}]-> should filter by edge properties."""
        db.execute_cypher(
            "CREATE (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}), "
            "(c:Person {name: 'Vincent'})"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (b:Person {name: 'Gus'}) "
            "CREATE (a)-[:KNOWS {since: 2020}]->(b)"
        )
        db.execute_cypher(
            "MATCH (a:Person {name: 'Alix'}), (c:Person {name: 'Vincent'}) "
            "CREATE (a)-[:KNOWS {since: 2023}]->(c)"
        )

        result = list(db.execute_cypher("MATCH (a)-[r {since: 2020}]->(b) RETURN a.name, b.name"))
        assert len(result) == 1
        assert result[0]["a.name"] == "Alix"
        assert result[0]["b.name"] == "Gus"



"""Cypher spec: Types (openCypher 9 sec 7).

Covers: Integer, Float, String, Boolean, Null, List, Map, Date, Time,
DateTime, Duration, Path, Node, Relationship. Spatial types (Point) are
marked xfail as they are not implemented.
"""

import pytest

# =============================================================================
# Scalar Types
# =============================================================================


class TestScalarTypes:
    """Integer, Float, String, Boolean, Null literals and type behavior."""

    def test_integer_decimal(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN 42 AS r"))
        assert result[0]["r"] == 42

    def test_integer_hex(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN 0xFF AS r"))
        assert result[0]["r"] == 255

    def test_integer_octal(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN 0o77 AS r"))
        assert result[0]["r"] == 63

    def test_float(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN 3.14 AS r"))
        assert abs(result[0]["r"] - 3.14) < 0.001

    def test_float_scientific(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN 1.5e2 AS r"))
        assert result[0]["r"] == 150.0

    def test_string_single_quoted(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN 'hello' AS r"))
        assert result[0]["r"] == "hello"

    def test_string_double_quoted(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher('MATCH (n:N) RETURN "hello" AS r'))
        assert result[0]["r"] == "hello"

    def test_boolean_true(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN true AS r"))
        assert result[0]["r"] is True

    def test_boolean_false(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN false AS r"))
        assert result[0]["r"] is False

    def test_null(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN null AS r"))
        assert result[0]["r"] is None


# =============================================================================
# Collection Types
# =============================================================================


class TestCollectionTypes:
    """List and Map types."""

    def test_list_heterogeneous(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN [1, 'two', 3.0, true] AS r"))
        r = result[0]["r"]
        assert r[0] == 1
        assert r[1] == "two"
        assert r[3] is True

    def test_map_type(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN {name: 'Alix', age: 30} AS r"))
        assert result[0]["r"]["name"] == "Alix"
        assert result[0]["r"]["age"] == 30


# =============================================================================
# Temporal Types
# =============================================================================


class TestTemporalTypes:
    """Date, Time, DateTime, Duration types."""

    def test_date_type(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN date('2024-01-15') AS r"))
        assert "2024-01-15" in str(result[0]["r"])

    def test_time_type(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN time('14:30:00') AS r"))
        assert "14:30" in str(result[0]["r"])

    def test_datetime_type(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN datetime('2024-01-15T14:30:00') AS r"))
        assert "2024" in str(result[0]["r"])

    def test_duration_type(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN duration('P1Y2M3D') AS r"))
        assert result[0]["r"] is not None


# =============================================================================
# Graph Element Types
# =============================================================================


class TestGraphElementTypes:
    """Node, Relationship, Path types."""

    def test_node_return(self, db):
        db.create_node(["Person"], {"name": "Alix"})
        result = list(db.execute_cypher("MATCH (n:Person) RETURN n"))
        assert len(result) == 1
        # Node returned as dict-like with properties
        assert result[0]["n"] is not None

    def test_relationship_return(self, db):
        a = db.create_node(["Person"], {"name": "Alix"})
        b = db.create_node(["Person"], {"name": "Gus"})
        db.create_edge(a.id, b.id, "KNOWS", {"since": 2020})
        result = list(db.execute_cypher("MATCH ()-[r:KNOWS]->() RETURN r"))
        assert len(result) == 1

    def test_path_return(self, db):
        a = db.create_node(["Person"], {"name": "Alix"})
        b = db.create_node(["Person"], {"name": "Gus"})
        db.create_edge(a.id, b.id, "KNOWS")
        result = list(db.execute_cypher("MATCH p = (:Person)-[:KNOWS]->(:Person) RETURN p"))
        assert len(result) == 1


# =============================================================================
# Spatial Types (MISSING)
# =============================================================================


class TestSpatialTypes:
    """Point types: not implemented."""

    @pytest.mark.xfail(reason="Spatial Point type not implemented")
    def test_point_2d_cartesian(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN point({x: 1.0, y: 2.0}) AS r"))
        assert result[0]["r"] is not None

    @pytest.mark.xfail(reason="Spatial Point type not implemented")
    def test_point_wgs84(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(
            db.execute_cypher("MATCH (n:N) RETURN point({latitude: 52.37, longitude: 4.89}) AS r")
        )
        assert result[0]["r"] is not None

    @pytest.mark.xfail(reason="Spatial Point type not implemented")
    def test_point_3d(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN point({x: 1.0, y: 2.0, z: 3.0}) AS r"))
        assert result[0]["r"] is not None



"""Cypher spec: Functions (openCypher 9 sec 6).

Covers: Aggregation, scalar, string, numeric, list, path, temporal functions.
"""

import math

import pytest

# =============================================================================
# Aggregation Functions (sec 6.1)
# =============================================================================


class TestAggregationFunctions:
    """count, sum, avg, min, max, collect, stDev, percentile."""

    def test_count_star(self, db):
        db.create_node(["N"], {"v": 1})
        db.create_node(["N"], {"v": 2})
        result = list(db.execute_cypher("MATCH (n:N) RETURN count(*) AS r"))
        assert result[0]["r"] == 2

    def test_count_expr(self, db):
        db.create_node(["N"], {"v": 1})
        db.create_node(["N"], {"v": 2})
        result = list(db.execute_cypher("MATCH (n:N) RETURN count(n) AS r"))
        assert result[0]["r"] == 2

    def test_count_distinct(self, db):
        db.create_node(["N"], {"v": "a"})
        db.create_node(["N"], {"v": "a"})
        db.create_node(["N"], {"v": "b"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN count(DISTINCT n.v) AS r"))
        assert result[0]["r"] == 2

    def test_sum(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 20})
        db.create_node(["N"], {"v": 30})
        result = list(db.execute_cypher("MATCH (n:N) RETURN sum(n.v) AS r"))
        assert result[0]["r"] == 60

    def test_avg(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 20})
        result = list(db.execute_cypher("MATCH (n:N) RETURN avg(n.v) AS r"))
        assert result[0]["r"] == 15.0

    def test_min(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 20})
        result = list(db.execute_cypher("MATCH (n:N) RETURN min(n.v) AS r"))
        assert result[0]["r"] == 5

    def test_max(self, db):
        db.create_node(["N"], {"v": 10})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 20})
        result = list(db.execute_cypher("MATCH (n:N) RETURN max(n.v) AS r"))
        assert result[0]["r"] == 20

    def test_collect(self, db):
        db.create_node(["N"], {"v": 1})
        db.create_node(["N"], {"v": 2})
        db.create_node(["N"], {"v": 3})
        result = list(db.execute_cypher("MATCH (n:N) RETURN collect(n.v) AS r ORDER BY r"))
        assert sorted(result[0]["r"]) == [1, 2, 3]

    def test_stdev(self, db):
        db.create_node(["N"], {"v": 2})
        db.create_node(["N"], {"v": 4})
        db.create_node(["N"], {"v": 4})
        db.create_node(["N"], {"v": 4})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 7})
        db.create_node(["N"], {"v": 9})
        result = list(db.execute_cypher("MATCH (n:N) RETURN stDev(n.v) AS r"))
        assert abs(result[0]["r"] - 2.0) < 0.5

    def test_stdevp(self, db):
        db.create_node(["N"], {"v": 2})
        db.create_node(["N"], {"v": 4})
        db.create_node(["N"], {"v": 4})
        db.create_node(["N"], {"v": 4})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 5})
        db.create_node(["N"], {"v": 7})
        db.create_node(["N"], {"v": 9})
        result = list(db.execute_cypher("MATCH (n:N) RETURN stDevP(n.v) AS r"))
        assert result[0]["r"] > 0

    def test_percentile_cont(self, db):
        for v in [1, 2, 3, 4, 5]:
            db.create_node(["N"], {"v": v})
        result = list(db.execute_cypher("MATCH (n:N) RETURN percentileCont(n.v, 0.5) AS r"))
        assert result[0]["r"] == 3.0

    def test_percentile_disc(self, db):
        for v in [1, 2, 3, 4, 5]:
            db.create_node(["N"], {"v": v})
        result = list(db.execute_cypher("MATCH (n:N) RETURN percentileDisc(n.v, 0.5) AS r"))
        assert result[0]["r"] == 3


# =============================================================================
# Scalar Functions (sec 6.2)
# =============================================================================


class TestScalarFunctions:
    """id, elementId, labels, type, keys, properties, exists, size, length."""

    def test_id(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN id(n) AS r"))
        assert isinstance(result[0]["r"], int)

    def test_element_id(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN elementId(n) AS r"))
        assert result[0]["r"] is not None

    def test_labels(self, db):
        db.create_node(["Person", "Developer"], {"name": "Alix"})
        result = list(db.execute_cypher("MATCH (n:Person) RETURN labels(n) AS r"))
        assert "Person" in result[0]["r"]
        assert "Developer" in result[0]["r"]

    def test_type(self, db):
        a = db.create_node(["Person"], {"name": "Alix"})
        b = db.create_node(["Person"], {"name": "Gus"})
        db.create_edge(a.id, b.id, "KNOWS")
        result = list(db.execute_cypher("MATCH ()-[r:KNOWS]->() RETURN type(r) AS r"))
        assert result[0]["r"] == "KNOWS"

    def test_keys(self, db):
        db.create_node(["N"], {"name": "Alix", "age": 30})
        result = list(db.execute_cypher("MATCH (n:N) RETURN keys(n) AS r"))
        assert "name" in result[0]["r"]
        assert "age" in result[0]["r"]

    def test_properties(self, db):
        db.create_node(["N"], {"name": "Alix", "age": 30})
        result = list(db.execute_cypher("MATCH (n:N) RETURN properties(n) AS r"))
        assert result[0]["r"]["name"] == "Alix"
        assert result[0]["r"]["age"] == 30

    def test_exists_property(self, db):
        """exists(n.prop) function form."""
        db.create_node(["N"], {"v": 1})
        db.create_node(["N"], {"w": 2})
        result = list(db.execute_cypher("MATCH (n:N) WHERE exists(n.v) RETURN count(n) AS cnt"))
        assert result[0]["cnt"] == 1

    def test_size_list(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN size([1, 2, 3]) AS r"))
        assert result[0]["r"] == 3

    def test_size_string(self, db):
        db.create_node(["N"], {"v": "hello"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN size(n.v) AS r"))
        assert result[0]["r"] == 5

    def test_length_string(self, db):
        db.create_node(["N"], {"v": "hello"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN length(n.v) AS r"))
        assert result[0]["r"] == 5

    def test_to_integer(self, db):
        db.create_node(["N"], {"v": "42"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN toInteger(n.v) AS r"))
        assert result[0]["r"] == 42

    def test_to_float(self, db):
        db.create_node(["N"], {"v": "3.14"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN toFloat(n.v) AS r"))
        assert abs(result[0]["r"] - 3.14) < 0.001

    def test_to_string(self, db):
        db.create_node(["N"], {"v": 42})
        result = list(db.execute_cypher("MATCH (n:N) RETURN toString(n.v) AS r"))
        assert result[0]["r"] == "42"

    def test_to_boolean(self, db):
        db.create_node(["N"], {"v": "true"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN toBoolean(n.v) AS r"))
        assert result[0]["r"] is True


# =============================================================================
# String Functions (sec 6.3)
# =============================================================================


class TestStringFunctions:
    """toUpper, toLower, trim, replace, substring, split, left, right, reverse."""

    def test_to_upper(self, db):
        db.create_node(["N"], {"v": "hello"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN toUpper(n.v) AS r"))
        assert result[0]["r"] == "HELLO"

    def test_to_lower(self, db):
        db.create_node(["N"], {"v": "HELLO"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN toLower(n.v) AS r"))
        assert result[0]["r"] == "hello"

    def test_trim(self, db):
        db.create_node(["N"], {"v": "  hello  "})
        result = list(db.execute_cypher("MATCH (n:N) RETURN trim(n.v) AS r"))
        assert result[0]["r"] == "hello"

    def test_ltrim(self, db):
        db.create_node(["N"], {"v": "  hello"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN ltrim(n.v) AS r"))
        assert result[0]["r"] == "hello"

    def test_rtrim(self, db):
        db.create_node(["N"], {"v": "hello  "})
        result = list(db.execute_cypher("MATCH (n:N) RETURN rtrim(n.v) AS r"))
        assert result[0]["r"] == "hello"

    def test_replace(self, db):
        db.create_node(["N"], {"v": "hello world"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN replace(n.v, 'world', 'grafeo') AS r"))
        assert result[0]["r"] == "hello grafeo"

    def test_substring(self, db):
        db.create_node(["N"], {"v": "Amsterdam"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN substring(n.v, 0, 5) AS r"))
        assert result[0]["r"] == "Amste"

    def test_split(self, db):
        db.create_node(["N"], {"v": "a,b,c"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN split(n.v, ',') AS r"))
        assert result[0]["r"] == ["a", "b", "c"]

    def test_left(self, db):
        db.create_node(["N"], {"v": "Amsterdam"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN left(n.v, 3) AS r"))
        assert result[0]["r"] == "Ams"

    def test_right(self, db):
        db.create_node(["N"], {"v": "Amsterdam"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN right(n.v, 3) AS r"))
        assert result[0]["r"] == "dam"

    def test_reverse(self, db):
        db.create_node(["N"], {"v": "hello"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN reverse(n.v) AS r"))
        assert result[0]["r"] == "olleh"

    def test_char_length(self, db):
        db.create_node(["N"], {"v": "hello"})
        result = list(db.execute_cypher("MATCH (n:N) RETURN char_length(n.v) AS r"))
        assert result[0]["r"] == 5


# =============================================================================
# Numeric Functions (sec 6.4)
# =============================================================================


class TestNumericFunctions:
    """abs, ceil, floor, round, sqrt, sign, log, exp, trig."""

    def test_abs(self, db):
        db.create_node(["N"], {"v": -5})
        result = list(db.execute_cypher("MATCH (n:N) RETURN abs(n.v) AS r"))
        assert result[0]["r"] == 5

    def test_ceil(self, db):
        db.create_node(["N"], {"v": 2.3})
        result = list(db.execute_cypher("MATCH (n:N) RETURN ceil(n.v) AS r"))
        assert result[0]["r"] == 3

    def test_floor(self, db):
        db.create_node(["N"], {"v": 2.7})
        result = list(db.execute_cypher("MATCH (n:N) RETURN floor(n.v) AS r"))
        assert result[0]["r"] == 2

    def test_round(self, db):
        db.create_node(["N"], {"v": 2.5})
        result = list(db.execute_cypher("MATCH (n:N) RETURN round(n.v) AS r"))
        assert result[0]["r"] == 3

    def test_sqrt(self, db):
        db.create_node(["N"], {"v": 9})
        result = list(db.execute_cypher("MATCH (n:N) RETURN sqrt(n.v) AS r"))
        assert result[0]["r"] == 3.0

    def test_sign(self, db):
        db.create_node(["N"], {"v": -5})
        result = list(db.execute_cypher("MATCH (n:N) RETURN sign(n.v) AS r"))
        assert result[0]["r"] == -1

    def test_log(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN log(e()) AS r"))
        assert abs(result[0]["r"] - 1.0) < 0.001

    def test_log10(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN log10(100) AS r"))
        assert abs(result[0]["r"] - 2.0) < 0.001

    def test_exp(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN exp(1) AS r"))
        assert abs(result[0]["r"] - math.e) < 0.001

    def test_e(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN e() AS r"))
        assert abs(result[0]["r"] - math.e) < 0.001

    def test_pi(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN pi() AS r"))
        assert abs(result[0]["r"] - math.pi) < 0.001

    def test_rand(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN rand() AS r"))
        assert 0.0 <= result[0]["r"] <= 1.0

    def test_sin(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN sin(0) AS r"))
        assert abs(result[0]["r"]) < 0.001

    def test_cos(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN cos(0) AS r"))
        assert abs(result[0]["r"] - 1.0) < 0.001

    def test_tan(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN tan(0) AS r"))
        assert abs(result[0]["r"]) < 0.001

    def test_asin(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN asin(1) AS r"))
        assert abs(result[0]["r"] - math.pi / 2) < 0.001

    def test_acos(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN acos(1) AS r"))
        assert abs(result[0]["r"]) < 0.001

    def test_atan(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN atan(1) AS r"))
        assert abs(result[0]["r"] - math.pi / 4) < 0.001

    def test_atan2(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN atan2(1, 1) AS r"))
        assert abs(result[0]["r"] - math.pi / 4) < 0.001

    def test_degrees(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN degrees(pi()) AS r"))
        assert abs(result[0]["r"] - 180.0) < 0.001

    def test_radians(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN radians(180) AS r"))
        assert abs(result[0]["r"] - math.pi) < 0.001


# =============================================================================
# List Functions (sec 6.5)
# =============================================================================


class TestListFunctions:
    """head, last, tail, size, range, reverse, reduce, keys."""

    def test_head(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN head([10, 20, 30]) AS r"))
        assert result[0]["r"] == 10

    def test_last(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN last([10, 20, 30]) AS r"))
        assert result[0]["r"] == 30

    def test_tail(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN tail([10, 20, 30]) AS r"))
        assert result[0]["r"] == [20, 30]

    def test_size(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN size([10, 20, 30]) AS r"))
        assert result[0]["r"] == 3

    def test_range(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN range(1, 5) AS r"))
        assert result[0]["r"] == [1, 2, 3, 4, 5]

    def test_range_with_step(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN range(0, 10, 3) AS r"))
        assert result[0]["r"] == [0, 3, 6, 9]

    def test_reverse_list(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN reverse([1, 2, 3]) AS r"))
        assert result[0]["r"] == [3, 2, 1]

    def test_keys_map(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN keys({a: 1, b: 2}) AS r"))
        assert sorted(result[0]["r"]) == ["a", "b"]


# =============================================================================
# Path Functions (sec 6.6)
# =============================================================================


class TestPathFunctions:
    """length, nodes, relationships on paths."""

    def test_length_path(self, db):
        a = db.create_node(["N"], {"v": 1})
        b = db.create_node(["N"], {"v": 2})
        c = db.create_node(["N"], {"v": 3})
        db.create_edge(a.id, b.id, "NEXT")
        db.create_edge(b.id, c.id, "NEXT")
        result = list(
            db.execute_cypher(
                "MATCH p = (a:N {v: 1})-[:NEXT*2]->(b:N {v: 3}) RETURN length(p) AS r"
            )
        )
        assert result[0]["r"] == 2

    def test_nodes_path(self, db):
        a = db.create_node(["N"], {"v": 1})
        b = db.create_node(["N"], {"v": 2})
        db.create_edge(a.id, b.id, "NEXT")
        result = list(
            db.execute_cypher("MATCH p = (a:N {v: 1})-[:NEXT]->(b) RETURN size(nodes(p)) AS r")
        )
        assert result[0]["r"] == 2

    def test_relationships_path(self, db):
        a = db.create_node(["N"], {"v": 1})
        b = db.create_node(["N"], {"v": 2})
        db.create_edge(a.id, b.id, "NEXT")
        result = list(
            db.execute_cypher(
                "MATCH p = (a:N {v: 1})-[:NEXT]->(b) RETURN size(relationships(p)) AS r"
            )
        )
        assert result[0]["r"] == 1


# =============================================================================
# Temporal Functions (sec 6.7)
# =============================================================================


class TestTemporalFunctions:
    """date, time, datetime, duration, now, temporal accessors."""

    def test_date_current(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN date() AS r"))
        assert result[0]["r"] is not None

    def test_date_parse(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN date('2024-01-15') AS r"))
        assert "2024-01-15" in str(result[0]["r"])

    def test_time_parse(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN time('14:30:00') AS r"))
        assert "14:30" in str(result[0]["r"])

    def test_datetime_parse(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN datetime('2024-01-15T14:30:00') AS r"))
        assert "2024" in str(result[0]["r"])

    def test_duration_parse(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN duration('P1Y2M') AS r"))
        assert result[0]["r"] is not None

    def test_now(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN now() AS r"))
        assert result[0]["r"] is not None

    def test_year_accessor(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN year(date('2024-03-15')) AS r"))
        assert result[0]["r"] == 2024

    def test_month_accessor(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN month(date('2024-03-15')) AS r"))
        assert result[0]["r"] == 3

    def test_day_accessor(self, db):
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN day(date('2024-03-15')) AS r"))
        assert result[0]["r"] == 15

    @pytest.mark.xfail(reason="Spatial types not implemented")
    def test_point_2d(self, db):
        """point({x, y}) 2D Cartesian."""
        db.create_node(["N"], {"v": 1})
        result = list(db.execute_cypher("MATCH (n:N) RETURN point({x: 1.0, y: 2.0}) AS r"))
        assert result[0]["r"] is not None

    @pytest.mark.xfail(reason="Spatial types not implemented")
    def test_distance(self, db):
        """distance(point1, point2)."""
        db.create_node(["N"], {"v": 1})
        result = list(
            db.execute_cypher(
                "MATCH (n:N) RETURN distance(point({x: 0, y: 0}), point({x: 3, y: 4})) AS r"
            )
        )
        assert result[0]["r"] == 5.0



"""SQL/PGQ-specific pytest fixtures and configuration."""

import pytest

# Try to import grafeo
try:
    from grafeo import GrafeoDB

    GRAFEO_AVAILABLE = True
except ImportError:
    GRAFEO_AVAILABLE = False


def has_sql_pgq_support(db):
    """Check if SQL/PGQ support is available."""
    try:
        db.execute_sql("SELECT * FROM GRAPH_TABLE (MATCH (n) COLUMNS (n.name AS name))")
        return True
    except (AttributeError, NotImplementedError):
        return False
    except Exception:
        return True  # Has method but might fail for other reasons


@pytest.fixture
def db():
    """Create a fresh in-memory GrafeoDB instance.

    This fixture requires SQL/PGQ support.
    """
    if not GRAFEO_AVAILABLE:
        pytest.skip("grafeo not installed")
    database = GrafeoDB()
    if not has_sql_pgq_support(database):
        pytest.skip("SQL/PGQ support not available in this build")
    return database


@pytest.fixture
def social_db(db):
    """Create a database with a social network for SQL/PGQ tests.

    Structure:
    - Alix (Person, age 30, city NYC) -KNOWS-> Gus (since 2020)
    - Gus (Person, age 25, city LA) -KNOWS-> Vincent (since 2021)
    - Alix -KNOWS-> Vincent (since 2019)
    - Vincent (Person, age 35, city NYC)
    - Acme Corp (Company, founded 2010)
    - Globex Inc (Company, founded 2015)
    - Alix -WORKS_AT-> Acme Corp (role Engineer)
    - Gus -WORKS_AT-> Globex Inc (role Manager)
    - Vincent -WORKS_AT-> Acme Corp (role Director)
    """
    alix = db.create_node(["Person"], {"name": "Alix", "age": 30, "city": "NYC"})
    gus = db.create_node(["Person"], {"name": "Gus", "age": 25, "city": "LA"})
    vincent = db.create_node(["Person"], {"name": "Vincent", "age": 35, "city": "NYC"})

    acme = db.create_node(["Company"], {"name": "Acme Corp", "founded": 2010})
    globex = db.create_node(["Company"], {"name": "Globex Inc", "founded": 2015})

    db.create_edge(alix.id, gus.id, "KNOWS", {"since": 2020})
    db.create_edge(gus.id, vincent.id, "KNOWS", {"since": 2021})
    db.create_edge(alix.id, vincent.id, "KNOWS", {"since": 2019})

    db.create_edge(alix.id, acme.id, "WORKS_AT", {"role": "Engineer"})
    db.create_edge(gus.id, globex.id, "WORKS_AT", {"role": "Manager"})
    db.create_edge(vincent.id, acme.id, "WORKS_AT", {"role": "Director"})

    return db


@pytest.fixture
def chain_db(db):
    """Create a chain graph for path tests: A -> B -> C -> D.

    All nodes are Person type with KNOWS edges.
    """
    a = db.create_node(["Person"], {"name": "A"})
    b = db.create_node(["Person"], {"name": "B"})
    c = db.create_node(["Person"], {"name": "C"})
    d = db.create_node(["Person"], {"name": "D"})

    db.create_edge(a.id, b.id, "KNOWS", {})
    db.create_edge(b.id, c.id, "KNOWS", {})
    db.create_edge(c.id, d.id, "KNOWS", {})

    return db



"""Base class for transaction tests.

This module defines test logic for transaction commit, rollback, and isolation.
"""

from abc import ABC, abstractmethod


class BaseTransactionsTest(ABC):
    """Abstract base class for transaction tests."""

    def execute_query(self, db, query):
        """Execute a query using the appropriate language parser.

        Override in subclasses that need a specific parser (e.g., Cypher).
        Default uses GQL parser via db.execute().
        """
        return db.execute(query)

    def execute_in_tx(self, tx, query):
        """Execute a query within a transaction.

        Override in subclasses that need a specific parser (e.g., Gremlin).
        Default uses tx.execute() (GQL parser).
        """
        return tx.execute(query)

    @abstractmethod
    def insert_query(self, labels: list[str], props: dict) -> str:
        """Return query to insert a node.

        Args:
            labels: Node labels
            props: Node properties

        Returns:
            Language-specific INSERT/CREATE query
        """
        raise NotImplementedError

    @abstractmethod
    def match_by_prop_query(self, label: str, prop: str, value) -> str:
        """Return query to match node by property.

        Args:
            label: Node label
            prop: Property name
            value: Property value

        Returns:
            Query that returns matching nodes
        """
        raise NotImplementedError

    @abstractmethod
    def count_query(self, label: str) -> str:
        """Return query to count nodes by label.

        Returns:
            Query that returns count as 'cnt'
        """
        raise NotImplementedError

    # ===== Test Methods =====

    def test_transaction_commit(self, db):
        """Test that committed transaction persists data."""
        with db.begin_transaction() as tx:
            query = self.insert_query(["Person"], {"name": "CommitTest"})
            self.execute_in_tx(tx, query)
            tx.commit()

        # Data should be visible after commit
        match_query = self.match_by_prop_query("Person", "name", "CommitTest")
        result = self.execute_query(db, match_query)
        rows = list(result)
        assert len(rows) == 1

    def test_transaction_auto_commit(self, db):
        """Test that transactions auto-commit on context exit."""
        with db.begin_transaction() as tx:
            query = self.insert_query(["Person"], {"name": "AutoCommitTest"})
            self.execute_in_tx(tx, query)
            # No explicit commit - should auto-commit

        # Data should be visible
        match_query = self.match_by_prop_query("Person", "name", "AutoCommitTest")
        result = self.execute_query(db, match_query)
        rows = list(result)
        assert len(rows) == 1

    def test_transaction_rollback(self, db):
        """Test that rollback discards changes."""
        # Verify database is empty for this label/prop combo
        match_query = self.match_by_prop_query("Person", "name", "RollbackTest")
        result = self.execute_query(db, match_query)
        assert len(list(result)) == 0

        # Create node and rollback
        with db.begin_transaction() as tx:
            query = self.insert_query(["Person"], {"name": "RollbackTest"})
            self.execute_in_tx(tx, query)
            tx.rollback()

        # Data should NOT be visible after rollback
        result = self.execute_query(db, match_query)
        rows = list(result)
        assert len(rows) == 0, f"Expected 0 rows after rollback, got {len(rows)}"

    def test_transaction_is_active(self, db):
        """Test transaction is_active property."""
        tx = db.begin_transaction()
        assert tx.is_active is True

        tx.commit()
        assert tx.is_active is False

    def test_multiple_operations_in_transaction(self, db):
        """Test multiple operations in a single transaction."""
        with db.begin_transaction() as tx:
            # Create multiple nodes
            self.execute_in_tx(tx, self.insert_query(["Person"], {"name": "Multi1", "idx": 1}))
            self.execute_in_tx(tx, self.insert_query(["Person"], {"name": "Multi2", "idx": 2}))
            self.execute_in_tx(tx, self.insert_query(["Person"], {"name": "Multi3", "idx": 3}))
            tx.commit()

        # All nodes should exist
        count_query = self.count_query("Person")
        result = self.execute_query(db, count_query)
        rows = list(result)
        # Get count value - column name varies by language (cnt, count, etc)
        # If count_query returns actual rows instead of a count, use len(rows)
        if len(rows) == 0:
            count_value = 0
        elif rows[0].get("cnt") is not None:
            count_value = rows[0].get("cnt")
        elif rows[0].get("count") is not None:
            count_value = rows[0].get("count")
        elif len(rows[0]) > 0:
            first_value = list(rows[0].values())[0]
            # If it's a number, use it as count; otherwise count the rows
            count_value = first_value if isinstance(first_value, (int, float)) else len(rows)
        else:
            count_value = len(rows)
        assert count_value >= 3



"""Cypher implementation of mutation tests.

Tests CRUD operations using Cypher query language.
"""

from tests.bases.test_mutations import BaseMutationsTest


class TestCypherMutations(BaseMutationsTest):
    """Cypher implementation of mutation tests."""

    def execute_query(self, db, query):
        """Execute query using Cypher parser."""
        return db.execute_cypher(query)

    def create_node_query(self, labels: list[str], props: dict) -> str:
        """Cypher: CREATE (:<labels> {<props>}) RETURN n"""
        label_str = ":".join(labels) if labels else ""
        if label_str:
            label_str = f":{label_str}"

        prop_parts = []
        for k, v in props.items():
            if isinstance(v, str):
                prop_parts.append(f"{k}: '{v}'")
            elif isinstance(v, bool):
                prop_parts.append(f"{k}: {'true' if v else 'false'}")
            elif v is None:
                prop_parts.append(f"{k}: null")
            else:
                prop_parts.append(f"{k}: {v}")

        prop_str = ", ".join(prop_parts)
        return f"CREATE (n{label_str} {{{prop_str}}}) RETURN n"

    def match_node_query(self, label: str, return_prop: str = "name") -> str:
        return f"MATCH (n:{label}) RETURN n.{return_prop}"

    def match_where_query(
        self, label: str, prop: str, op: str, value, return_prop: str = "name"
    ) -> str:
        value_str = f"'{value}'" if isinstance(value, str) else str(value)
        return f"MATCH (n:{label}) WHERE n.{prop} {op} {value_str} RETURN n.{return_prop}"

    def delete_node_query(self, label: str, prop: str, value) -> str:
        value_str = f"'{value}'" if isinstance(value, str) else str(value)
        return f"MATCH (n:{label}) WHERE n.{prop} = {value_str} DELETE n"

    def create_edge_query(
        self,
        from_label: str,
        from_prop: str,
        from_value,
        to_label: str,
        to_prop: str,
        to_value,
        edge_type: str,
        edge_props: dict,
    ) -> str:
        from_val = f"'{from_value}'" if isinstance(from_value, str) else from_value
        to_val = f"'{to_value}'" if isinstance(to_value, str) else to_value

        prop_parts = []
        for k, v in edge_props.items():
            if isinstance(v, str):
                prop_parts.append(f"{k}: '{v}'")
            else:
                prop_parts.append(f"{k}: {v}")
        prop_str = ", ".join(prop_parts) if prop_parts else ""

        if prop_str:
            return (
                f"MATCH (a:{from_label}), (b:{to_label}) "
                f"WHERE a.{from_prop} = {from_val} AND b.{to_prop} = {to_val} "
                f"CREATE (a)-[r:{edge_type} {{{prop_str}}}]->(b) RETURN r"
            )
        else:
            return (
                f"MATCH (a:{from_label}), (b:{to_label}) "
                f"WHERE a.{from_prop} = {from_val} AND b.{to_prop} = {to_val} "
                f"CREATE (a)-[r:{edge_type}]->(b) RETURN r"
            )

    def update_node_query(
        self, label: str, match_prop: str, match_value, set_prop: str, set_value
    ) -> str:
        match_val = f"'{match_value}'" if isinstance(match_value, str) else match_value
        set_val = f"'{set_value}'" if isinstance(set_value, str) else set_value
        return (
            f"MATCH (n:{label}) WHERE n.{match_prop} = {match_val} "
            f"SET n.{set_prop} = {set_val} RETURN n"
        )


# =============================================================================
# CYPHER-SPECIFIC MUTATION TESTS
# =============================================================================


class TestCypherSpecificMutations:
    """Cypher-specific mutation tests."""

    def test_cypher_create_syntax(self, db):
        """Test Cypher CREATE syntax."""
        result = db.execute_cypher("CREATE (n:Person {name: 'CreateTest', age: 42}) RETURN n")
        rows = list(result)
        assert len(rows) == 1

        result = db.execute_cypher("MATCH (n:Person) WHERE n.name = 'CreateTest' RETURN n.age")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.age"] == 42

    def test_cypher_merge(self, db):
        """Test Cypher MERGE (create if not exists)."""
        db.execute_cypher("MERGE (c:City {name: 'NYC'}) RETURN c")
        db.execute_cypher("MERGE (c:City {name: 'NYC'}) RETURN c")

        result = db.execute_cypher("MATCH (c:City) RETURN count(c) AS cnt")
        rows = list(result)
        assert rows[0]["cnt"] == 1

    def test_cypher_set(self, db):
        """Test Cypher SET for property update."""
        db.execute_cypher("CREATE (p:Person {name: 'SetTest', verified: false})")
        db.execute_cypher("MATCH (p:Person {name: 'SetTest'}) SET p.verified = true")

        result = db.execute_cypher("MATCH (p:Person {name: 'SetTest'}) RETURN p.verified")
        rows = list(result)
        assert rows[0]["p.verified"] is True

    def test_cypher_set_add_property(self, db):
        """Test Cypher SET to add new property."""
        db.execute_cypher("CREATE (p:Person {name: 'AddProp'})")
        db.execute_cypher("MATCH (p:Person {name: 'AddProp'}) SET p.newProp = 'added'")

        result = db.execute_cypher("MATCH (p:Person {name: 'AddProp'}) RETURN p.newProp")
        rows = list(result)
        assert rows[0]["p.newProp"] == "added"

    def test_cypher_remove(self, db):
        """Test Cypher REMOVE property."""
        db.execute_cypher("CREATE (p:Person {name: 'RemoveTest', toRemove: 'value'})")
        db.execute_cypher("MATCH (p:Person {name: 'RemoveTest'}) REMOVE p.toRemove")

        result = db.execute_cypher("MATCH (p:Person {name: 'RemoveTest'}) RETURN p.toRemove")
        rows = list(result)
        assert rows[0].get("p.toRemove") is None

    def test_cypher_detach_delete(self, db):
        """Test Cypher DETACH DELETE (deletes node and all relationships)."""
        db.execute_cypher("CREATE (a:Node {name: 'A'})")
        db.execute_cypher("CREATE (b:Node {name: 'B'})")
        db.execute_cypher(
            "MATCH (a:Node {name: 'A'}), (b:Node {name: 'B'}) CREATE (a)-[:CONNECTED]->(b)"
        )
        db.execute_cypher("MATCH (n:Node {name: 'A'}) DETACH DELETE n")

        result = db.execute_cypher("MATCH (n:Node {name: 'A'}) RETURN n")
        rows = list(result)
        assert len(rows) == 0

        result = db.execute_cypher("MATCH (n:Node {name: 'B'}) RETURN n")
        rows = list(result)
        assert len(rows) == 1

    def test_cypher_create_negative_float(self, db):
        """Test CREATE with negative float property (issue #160)."""
        db.execute_cypher("CREATE (:Point {lat: -6.248, lon: 106.845})")

        result = db.execute_cypher("MATCH (n:Point) RETURN n.lat, n.lon")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.lat"] == -6.248
        assert rows[0]["n.lon"] == 106.845

    def test_cypher_create_negative_integer(self, db):
        """Test CREATE with negative integer property (issue #160)."""
        db.execute_cypher("CREATE (:Floor {level: -3})")

        result = db.execute_cypher("MATCH (n:Floor) RETURN n.level")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.level"] == -3

    def test_cypher_create_multiple_negative_properties(self, db):
        """Test CREATE with multiple negative properties (issue #160)."""
        db.execute_cypher("CREATE (:Point {x: -1, y: -2.5, z: -3})")

        result = db.execute_cypher("MATCH (n:Point) RETURN n.x, n.y, n.z")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.x"] == -1
        assert rows[0]["n.y"] == -2.5
        assert rows[0]["n.z"] == -3

    def test_cypher_create_edge_with_negative_property(self, db):
        """Test CREATE edge with negative property (issue #160)."""
        db.execute_cypher(
            "CREATE (:Node {name: 'Alix'})-[:LINK {weight: -0.75}]->(:Node {name: 'Gus'})"
        )

        result = db.execute_cypher("MATCH ()-[r:LINK]->() RETURN r.weight")
        rows = list(result)
        assert rows[0]["r.weight"] == -0.75

    def test_cypher_set_negative_value(self, db):
        """Test Cypher SET with negative value (issue #160)."""
        db.execute_cypher("CREATE (:Sensor {name: 'Temp', reading: 25})")
        db.execute_cypher("MATCH (n:Sensor {name: 'Temp'}) SET n.reading = -18.5")

        result = db.execute_cypher("MATCH (n:Sensor {name: 'Temp'}) RETURN n.reading")
        rows = list(result)
        assert rows[0]["n.reading"] == -18.5

    def test_cypher_where_with_negative(self, db):
        """Test Cypher WHERE with negative comparison (issue #160)."""
        db.execute_cypher("CREATE (:Floor {level: -1, name: 'B1'})")
        db.execute_cypher("CREATE (:Floor {level: -2, name: 'B2'})")
        db.execute_cypher("CREATE (:Floor {level: 0, name: 'G'})")

        result = db.execute_cypher("MATCH (n:Floor) WHERE n.level = -2 RETURN n.name")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.name"] == "B2"

    def test_cypher_merge_with_negative(self, db):
        """Test Cypher MERGE with negative property (issue #160)."""
        db.execute_cypher("MERGE (:Floor {level: -1, name: 'Basement'})")

        result = db.execute_cypher("MATCH (n:Floor) RETURN n.level, n.name")
        rows = list(result)
        assert len(rows) == 1
        assert rows[0]["n.level"] == -1
        assert rows[0]["n.name"] == "Basement"



"""Root pytest fixtures for Grafeo Python tests.

This module provides common fixtures available to all test modules.
"""

import sys
from pathlib import Path

# Add the python binding root to sys.path so imports like 'tests.bases' work
python_binding_root = Path(__file__).parent.parent
if str(python_binding_root) not in sys.path:
    sys.path.insert(0, str(python_binding_root))

import random  # noqa: E402

import pytest  # noqa: E402

# Try to import grafeo
try:
    import grafeo

    GRAFEO_AVAILABLE = True
except ImportError:
    GRAFEO_AVAILABLE = False


# Import fixtures from fixtures module
try:
    from tests.fixtures.datasets import (  # noqa: E402
        create_social_graph,
    )
    from tests.fixtures.generators import (  # noqa: E402
        SocialNetworkGenerator,
        load_data_into_db,
    )
except ImportError:
    # Fallback if relative imports don't work
    pass


@pytest.fixture
def db():
    """Create a fresh in-memory GrafeoDB instance."""
    if not GRAFEO_AVAILABLE:
        pytest.skip("grafeo not installed")
    return grafeo.GrafeoDB()


@pytest.fixture
def node_ids(db):
    """Create a test graph and return node IDs.

    Creates a random graph with 100 nodes and 300 edges for algorithm testing.
    """
    n_nodes = 100
    n_edges = 300

    node_ids = []
    for i in range(n_nodes):
        node = db.create_node(["Node"], {"index": i})
        node_ids.append(node.id)

    edges = set()
    while len(edges) < n_edges:
        src = random.choice(node_ids)
        dst = random.choice(node_ids)
        if src != dst and (src, dst) not in edges:
            db.create_edge(src, dst, "EDGE", {"weight": random.uniform(0.1, 10.0)})
            edges.add((src, dst))

    return node_ids


@pytest.fixture
def social_graph(db):
    """Create a social network graph.

    Creates 50 Person nodes with KNOWS edges.
    """
    if not GRAFEO_AVAILABLE:
        pytest.skip("grafeo not installed")

    try:
        return create_social_graph(db, size=50)
    except NameError:
        # Fallback implementation
        gen = SocialNetworkGenerator(num_nodes=50, avg_edges_per_node=5, seed=42)
        node_count, edge_count = load_data_into_db(db, gen)
        return {"node_count": node_count, "edge_count": edge_count}


@pytest.fixture
def pattern_graph(db):
    """Create a graph for pattern matching tests.

    Creates Person and Company nodes with various relationships.
    """
    if not GRAFEO_AVAILABLE:
        pytest.skip("grafeo not installed")

    # Create Person nodes
    alix = db.create_node(["Person"], {"name": "Alix", "age": 30, "city": "NYC"})
    gus = db.create_node(["Person"], {"name": "Gus", "age": 25, "city": "LA"})
    vincent = db.create_node(["Person"], {"name": "Vincent", "age": 35, "city": "NYC"})

    # Create Company nodes
    acme = db.create_node(["Company"], {"name": "Acme Corp", "founded": 2010})
    globex = db.create_node(["Company"], {"name": "Globex Inc", "founded": 2015})

    # Create KNOWS edges
    db.create_edge(alix.id, gus.id, "KNOWS", {"since": 2020})
    db.create_edge(gus.id, vincent.id, "KNOWS", {"since": 2021})
    db.create_edge(alix.id, vincent.id, "KNOWS", {"since": 2019})

    # Create WORKS_AT edges
    db.create_edge(alix.id, acme.id, "WORKS_AT", {"role": "Engineer"})
    db.create_edge(gus.id, globex.id, "WORKS_AT", {"role": "Manager"})
    db.create_edge(vincent.id, acme.id, "WORKS_AT", {"role": "Director"})

    return {
        "alix": alix,
        "gus": gus,
        "vincent": vincent,
        "acme": acme,
        "globex": globex,
    }


# Register markers
def pytest_configure(config):
    """Register custom markers."""
    config.addinivalue_line(
        "markers", "slow: marks tests as slow (deselect with '-m \"not slow\"')"
    )
    config.addinivalue_line("markers", "benchmark: marks tests as benchmarks")
    config.addinivalue_line("markers", "gql: marks tests requiring GQL support")
    config.addinivalue_line("markers", "cypher: marks tests requiring Cypher support")
    config.addinivalue_line("markers", "gremlin: marks tests requiring Gremlin support")
    config.addinivalue_line("markers", "graphql: marks tests requiring GraphQL support")
    config.addinivalue_line("markers", "sparql: marks tests requiring SPARQL support")



"""Cypher-specific pytest fixtures and configuration."""

import os

import pytest
from grafeo import GrafeoDB

_db = GrafeoDB()
HAS_CYPHER = hasattr(_db, "execute_cypher")

_THIS_DIR = os.path.dirname(__file__)


def pytest_collection_modifyitems(config, items):
    """Skip Cypher tests when the feature is not compiled in."""
    if HAS_CYPHER:
        return
    skip = pytest.mark.skip(reason="grafeo built without cypher feature")
    for item in items:
        if str(item.fspath).startswith(_THIS_DIR):
            item.add_marker(skip)



"""GQL hybrid search integration tests."""

import pytest

try:
    from grafeo import GrafeoDB

    GRAFEO_AVAILABLE = True
except ImportError:
    GRAFEO_AVAILABLE = False


@pytest.fixture
def db():
    if not GRAFEO_AVAILABLE:
        pytest.skip("grafeo not installed")
    return GrafeoDB()


@pytest.fixture
def hybrid_db(db):
    """Database with both text and vector indexes."""
    db.create_node(
        ["Doc"],
        {"content": "Rust graph database engine", "emb": [1.0, 0.0, 0.0]},
    )
    db.create_node(
        ["Doc"],
        {"content": "Python machine learning", "emb": [0.0, 1.0, 0.0]},
    )
    db.create_node(
        ["Doc"],
        {"content": "Rust systems programming", "emb": [0.9, 0.1, 0.0]},
    )
    db.create_node(
        ["Doc"],
        {"content": "Graph neural network", "emb": [0.5, 0.5, 0.0]},
    )

    db.create_text_index("Doc", "content")
    db.create_vector_index("Doc", "emb", dimensions=3, metric="cosine")
    return db


class TestHybridSearch:
    def test_hybrid_search_basic(self, hybrid_db):
        results = hybrid_db.hybrid_search(
            "Doc",
            text_property="content",
            vector_property="emb",
            query_text="Rust graph",
            k=4,
            query_vector=[1.0, 0.0, 0.0],
        )
        assert len(results) > 0

    def test_hybrid_search_text_only(self, hybrid_db):
        results = hybrid_db.hybrid_search(
            "Doc",
            text_property="content",
            vector_property="emb",
            query_text="Rust",
            k=4,
        )
        assert len(results) > 0

    def test_hybrid_search_no_text_matches(self, hybrid_db):
        # Even with no text matches, vector search may contribute
        results = hybrid_db.hybrid_search(
            "Doc",
            text_property="content",
            vector_property="emb",
            query_text="nonexistentxyzquery",
            k=4,
            query_vector=[0.0, 0.0, 0.0],
        )
        # Should not error
        assert isinstance(results, list)
